from pydantic import BaseModel, Field

from ..data_models import OllamaEndpointResponseBody
from .message_models import Message


class Function(BaseModel):
    name: str = Field(None, description="The name of the function to call.")
    arguments: str = Field(None, description="The arguments to pass to the function.")


class ToolCall(BaseModel):
    function: Function = Field(None, description="The function call details.")


class OllamaStreamChatCompletionResponseBody(OllamaEndpointResponseBody):
    model: str = Field(None, description="The model name")

    created_at: str = Field(
        None, description="The timestamp when the response was created"
    )

    message: Message = Field(
        None, description="he partial or full response generated by the model"
    )

    done: bool = Field(
        None,
        description="A flag indicating whether the response generation is complete",
    )

    done_reason: str = Field(
        None, description="The response generation complete reason"
    )


class OllamaChatCompletionResponseBody(OllamaStreamChatCompletionResponseBody):
    total_duration: int = Field(None, description="Time spent generating the response")

    load_duration: int = Field(
        None, description="Time spent in nanoseconds loading the model"
    )

    prompt_eval_count: int = Field(None, description="number of tokens in the prompt")

    prompt_eval_duration: int = Field(
        None, description="time spent in nanoseconds evaluating the prompt"
    )

    eval_count: int = Field(None, description="number of tokens in the response")

    eval_duration: int = Field(
        None, description="time in nanoseconds spent generating the response"
    )



from typing import Optional

from pydantic import Field

from ..data_models import OllamaEndpointResponseBody


class OllamaStreamCompletionResponseBody(OllamaEndpointResponseBody):
    model: str = Field(None, description="The model name")

    created_at: str = Field(
        None, description="The timestamp when the response was created"
    )

    response: str = Field(
        None, description="he partial or full response generated by the model"
    )

    done: bool = Field(
        None,
        description="A flag indicating whether the response generation is complete",
    )


class OllamaCompletionResponseBody(OllamaStreamCompletionResponseBody):
    total_duration: int = Field(None, description="Time spent generating the response")

    load_duration: int = Field(
        None, description="Time spent in nanoseconds loading the model"
    )

    prompt_eval_count: int = Field(None, description="number of tokens in the prompt")

    prompt_eval_duration: int = Field(
        None, description="time spent in nanoseconds evaluating the prompt"
    )

    eval_count: int = Field(None, description="number of tokens in the response")

    eval_duration: int = Field(
        None, description="time in nanoseconds spent generating the response"
    )

    context: list | None = Field(
        None,
        description="an encoding of the conversation used in this response, "
        "this can be sent in the next request to keep a conversational memory",
    )



import inspect

from dotenv import load_dotenv

from lion.integrations.perplexity_.api_endpoints.api_request import PerplexityRequest
from lion.integrations.perplexity_.api_endpoints.chat_completions.request.request_body import (
    PerplexityChatCompletionRequestBody,
)
from lion.service import Service, register_service

from .PerplexityModel import PerplexityModel

load_dotenv()


@register_service
class PerplexityService(Service):
    def __init__(
        self,
        api_key: str,
        name: str = None,
    ):
        super().__setattr__("_initialized", False)
        self.api_key = api_key
        self.name = name
        self.rate_limiters = {}  # model: RateLimiter
        super().__setattr__("_initialized", True)

    def __setattr__(self, key, value):
        if getattr(self, "_initialized", False) and key in [
            "api_key",
        ]:
            raise AttributeError(
                f"Cannot modify '{key}' after initialization. "
                f"Please set a new service object for new keys."
            )
        super().__setattr__(key, value)

    def check_rate_limiter(
        self,
        perplexity_model: PerplexityModel,
        limit_requests: int = None,
        limit_tokens: int = None,
    ):
        # Map model versions to their base models for shared rate limiting
        shared_models = {
            "llama-3.1-sonar-small-128k-online": "llama-3.1-sonar-small",
            "llama-3.1-sonar-medium-128k-online": "llama-3.1-sonar-medium",
            "llama-3.1-sonar-large-128k-online": "llama-3.1-sonar-large",
        }

        if perplexity_model.model in shared_models:
            model = shared_models[perplexity_model.model]
        else:
            model = perplexity_model.model

        if model not in self.rate_limiters:
            self.rate_limiters[model] = perplexity_model.rate_limiter
        else:
            perplexity_model.rate_limiter = self.rate_limiters[model]
            if limit_requests:
                perplexity_model.rate_limiter.limit_requests = limit_requests
            if limit_tokens:
                perplexity_model.rate_limiter.limit_tokens = limit_tokens

        return perplexity_model

    @staticmethod
    def match_data_model(task_name: str) -> dict:
        """Match task name to appropriate request and response models."""
        if task_name == "create_chat_completion":
            return {"request_body": PerplexityChatCompletionRequestBody}
        raise ValueError(f"No data models found for task: {task_name}")

    @classmethod
    def list_tasks(cls):
        methods = []
        for name, member in inspect.getmembers(cls, predicate=inspect.isfunction):
            if name not in [
                "__init__",
                "__setattr__",
                "check_rate_limiter",
                "match_data_model",
            ]:
                methods.append(name)
        return methods

    # Chat Completions
    def create_chat_completion(
        self, model: str, limit_tokens: int = None, limit_requests: int = None
    ):
        model_obj = PerplexityModel(
            model=model,
            api_key=self.api_key,
            endpoint="chat/completions",
            method="POST",
            content_type="application/json",
            limit_tokens=limit_tokens,
            limit_requests=limit_requests,
        )

        return self.check_rate_limiter(
            model_obj, limit_requests=limit_requests, limit_tokens=limit_tokens
        )



from pydantic import Field

from ..data_models import OpenAIEndpointPathParam, OpenAIEndpointResponseBody


class OpenAIDeleteFineTunedModelPathParam(OpenAIEndpointPathParam):
    model: str = Field(description="The model to delete.")


class OpenAIDeleteFineTunedModelResponseBody(OpenAIEndpointResponseBody):
    id: str = Field(
        description="The file identifier, which can be referenced in the API endpoints.",
    )

    object: str = Field(description="The object type, which is always model.")

    deleted: bool = Field(description="The deletion status")



from .request.request_body import OpenAIChatCompletionRequestBody


def get_text_messages(request_body: OpenAIChatCompletionRequestBody):
    messages_list = request_body.model_dump(exclude_unset=True).get("messages")
    parsed_str = "["

    for msg in messages_list:
        role = msg.get("role", "")
        content = msg.get("content", "")

        if isinstance(content, list):  # Check if content is a list (second example)
            content_str = []
            for sub_content in content:
                if sub_content.get("type") == "text":
                    content_str.append(f'{sub_content.get("text", "")}')
            content = " ".join(
                content_str
            )  # Combine all sub_content items into a single string
        parsed_str += f"role: {role} content: {content} "

    parsed_str = parsed_str.strip()  # Remove trailing space
    parsed_str += "]"

    return parsed_str


def get_images(request_body: OpenAIChatCompletionRequestBody):
    messages_list = request_body.model_dump(exclude_unset=True).get("messages")
    image_urls = []

    for msg in messages_list:
        content = msg.get("content", "")

        if isinstance(content, list):  # Check if content is a list (second example)
            for sub_content in content:
                if sub_content.get("type") == "image_url":
                    image_url = sub_content.get("image_url")
                    url = image_url.get("url")
                    detail = image_url.get("detail", "auto")
                    image_urls.append((url, detail))
    return image_urls



from typing import Literal

from pydantic import BaseModel, ConfigDict, Field, field_serializer, model_validator

from ...data_models import AnthropicEndpointResponseBody
from .content_models import ContentBlock
from .usage_models import Usage


class AnthropicMessageResponseBody(AnthropicEndpointResponseBody):
    id: str = Field(description="Unique object identifier")

    type: Literal[
        "message",
        "message_start",
        "content_block_start",
        "content_block_delta",
        "content_block_stop",
        "message_delta",
        "message_stop",
    ] = Field(description="Type of the response")

    role: Literal["assistant"] | None = Field(
        default=None,
        description="Conversational role of the generated message. Always 'assistant' for non-stream responses",
    )

    content: list[ContentBlock] | None = Field(
        default=None,
        description="Content generated by the model",
    )

    model: str | None = Field(
        default=None,
        description="The model that handled the request",
    )

    stop_reason: (
        Literal["end_turn", "max_tokens", "stop_sequence", "tool_use"] | None
    ) = Field(
        default=None,
        description="""The reason that we stopped. May be:
        "end_turn": the model reached a natural stopping point
        "max_tokens": exceeded the requested max_tokens or model's maximum
        "stop_sequence": one of the provided custom stop_sequences was generated
        "tool_use": the model invoked one or more tools""",
    )

    stop_sequence: str | None = Field(
        default=None,
        description="Which custom stop sequence was generated, if any",
    )

    usage: Usage | None = Field(
        default=None,
        description="""Billing and rate-limit usage. Anthropic's API bills and
        rate-limits by token counts, as tokens represent the underlying cost to our systems.""",
    )

    # Stream-specific fields
    message: "AnthropicMessageResponseBody | None" = Field(
        default=None,
        description="Initial message object in stream response",
    )

    content_block: dict | None = Field(
        default=None,
        description="Content block information in stream response",
    )

    delta: dict | None = Field(
        default=None,
        description="Delta update in stream response",
    )

    message_id: str | None = Field(
        default=None,
        description="Message ID reference in stream response",
    )

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @model_validator(mode="after")
    def validate_stream_fields(self):
        # For stream responses, some fields are optional
        if self.type != "message":
            return self

        # For regular message responses, validate required fields
        if not self.role:
            raise ValueError("role is required for message responses")
        if self.content is None:
            raise ValueError("content is required for message responses")
        if not self.model:
            raise ValueError("model is required for message responses")

        return self



from typing import List, Literal, Optional, Union

from pydantic import ConfigDict, Field, model_validator

from ..data_models import OpenAIEndpointRequestBody

InputType = Union[str, list[str], list[int], list[list[int]]]
EncodingFormat = Literal["float", "base64"]


class OpenAIEmbeddingRequestBody(OpenAIEndpointRequestBody):
    input: InputType = Field(
        description=(
            "Input text to embed, encoded as a string or array of tokens. "
            "To embed multiple inputs in a single request, pass an array of "
            "strings or array of token arrays. The input must not exceed "
            "the max input tokens for the model (8192 tokens for "
            "text-embedding-ada-002), cannot be an empty string, and any "
            "array must be 2048 dimensions or less."
        ),
    )

    model: str = Field(
        description=(
            "ID of the model to use. You can use the List models API to see "
            "all of your available models, or see our Model overview for "
            "descriptions of them."
        ),
    )

    encoding_format: EncodingFormat | None = Field(
        "float",
        description=(
            "The format to return the embeddings in. Can be either `float` "
            "or `base64`."
        ),
    )

    dimensions: int | None = Field(
        None,
        description=(
            "The number of dimensions the resulting output embeddings "
            "should have. Only supported in `text-embedding-3` and later "
            "models."
        ),
    )

    user: str | None = Field(
        None,
        description=(
            "A unique identifier representing your end-user, which can help "
            "OpenAI to monitor and detect abuse."
        ),
    )

    @model_validator(mode="after")
    def validate_input(self) -> "OpenAIEmbeddingRequestBody":
        if isinstance(self.input, str) and self.input.strip() == "":
            raise ValueError("Input cannot be an empty string.")
        if isinstance(self.input, list):
            if len(self.input) == 0:
                raise ValueError("Input array cannot be empty.")
            if isinstance(self.input[0], list) and len(self.input) > 2048:
                raise ValueError("Input array must be 2048 dimensions or less.")
        return self

    model_config = ConfigDict(
        json_schema_extra={
            "example": {
                "input": "The food was delicious and the waiter...",
                "model": "text-embedding-ada-002",
                "encoding_format": "float",
                "dimensions": None,
                "user": "user123",
            }
        }
    )



from typing import List, Literal, Optional

from pydantic import Field

from ..data_models import OpenAIEndpointQueryParam, OpenAIEndpointResponseBody
from .file_models import OpenAIFileResponseBody


class OpenAIListFilesQueryParam(OpenAIEndpointQueryParam):
    purpose: str | None = Field(description="Only return files with the given purpose.")


class OpenAIListFilesResponseBody(OpenAIEndpointResponseBody):
    data: list[OpenAIFileResponseBody] = Field(description="The list of file objects.")

    object: Literal["list"] = Field(
        description='The object type, which is always "list".'
    )

    has_more: bool = Field(
        description="Whether there are more results" " available after this batch."
    )



import re
from collections.abc import Callable
from typing import Any

from ..type_convert.to_dict import to_dict


def extract_json_blocks(
    str_to_parse: str,
    suppress: bool = True,
    fuzzy_parse: bool = True,
    dropna: bool = True,
) -> list[dict[str, Any]]:
    """
    Extract and parse JSON blocks from the given text.

    This function searches for JSON blocks enclosed in triple backticks
    within the input text, parses them, and returns a list of parsed
    dictionaries.

    Args:
        text: The input text containing JSON blocks.
        suppress: If True, suppress errors during parsing. Default is True.
        fuzzy_parse: If True, use fuzzy parsing for JSON. Default is True.
        dropna: If True, remove None values from the result. Default is True.

    Returns:
        A list of parsed JSON blocks as dictionaries.

    Example:
        >>> text = "```json\n{\"key\": \"value\"}\n```"
        >>> extract_json_blocks(text)
        [{'key': 'value'}]
    """
    pattern = r"```json\s*(.*?)\s*```"
    matches = re.findall(pattern, str_to_parse, re.DOTALL)

    json_blocks = [
        to_dict(match, fuzzy_parse=fuzzy_parse, suppress=suppress) for match in matches
    ]

    return [block for block in json_blocks if block] if dropna else json_blocks


def extract_block(
    str_to_parse: str,
    language: str = "json",
    regex_pattern: str | None = None,
    *,
    parser: Callable[[str], Any] | None = None,
    suppress: bool = False,
) -> dict[str, Any] | None:
    """
    Extract and parse a code block from the given string.

    This function searches for a code block in the input string using a
    regular expression pattern, extracts it, and parses it using the
    provided parser function.

    Args:
        str_to_parse: The input string containing the code block.
        language: The language of the code block. Default is "json".
        regex_pattern: Custom regex pattern to find the code block.
            If provided, overrides the default pattern.
        parser: A function to parse the extracted code string.
            If not provided, uses `to_dict` with fuzzy parsing.
        suppress: If True, return None instead of raising an error
            when no code block is found. Default is False.

    Returns:
        The parsed content of the code block as a dictionary,
        or None if no block is found and suppress is True.

    Raises:
        ValueError: If no code block is found and suppress is False.

    Example:
        >>> text = "```json\n{\"key\": \"value\"}\n```"
        >>> extract_block(text)
        {'key': 'value'}
    """
    if not regex_pattern:
        regex_pattern = rf"```{language}\n?(.*?)\n?```"
    if not language:
        regex_pattern = r"```\n?(.*?)\n?```"

    match = re.search(regex_pattern, str_to_parse, re.DOTALL)

    if match:
        code_str = match.group(1).strip()
    elif str_to_parse.startswith(f"```{language}\n") and str_to_parse.endswith("\n```"):
        code_str = str_to_parse[4 + len(language) : -4].strip()
    elif suppress:
        return None
    else:
        raise ValueError("No code block found in the input string.")

    parser = parser or (lambda x: to_dict(x, fuzzy_parse=True, suppress=True))
    return parser(code_str)



"""Utility functions for handling responses."""

from typing import Any, Dict, List, Union

from .data_models import GroqAudioResponse, GroqChatCompletionResponse


def match_response(
    request_model: Any,
    response: dict[str, Any] | list[dict[str, Any]] | None,
) -> dict[str, Any] | list[dict[str, Any]] | None:
    """Match response to appropriate model and format."""
    if response is None:
        return None

    endpoint = request_model.endpoint.split("/")[0]

    # Chat completions endpoint
    if endpoint == "chat":
        if isinstance(response, dict):
            # Single response
            try:
                # Validate and parse through Pydantic model
                parsed = GroqChatCompletionResponse(**response)
                # Return in consistent format
                return {
                    "choices": parsed.choices,
                    "model": parsed.model,
                    "usage": parsed.usage,
                }
            except Exception as e:
                import warnings

                warnings.warn(f"Failed to parse chat response: {str(e)}")
                return response
        else:
            # Stream response list
            result = []
            for chunk in response:
                if not isinstance(chunk, dict):
                    continue

                try:
                    if "choices" in chunk:
                        # Regular chunk with content
                        result.append(
                            {
                                "choices": [
                                    {
                                        "delta": {
                                            "content": choice.get("delta", {}).get(
                                                "content", ""
                                            ),
                                            "role": "assistant",
                                        }
                                    }
                                    for choice in chunk["choices"]
                                ]
                            }
                        )

                        # Add usage if present
                        if "usage" in chunk:
                            result[-1]["usage"] = chunk["usage"]

                    elif "usage" in chunk:
                        # Final chunk with usage stats
                        result.append(
                            {
                                "choices": [
                                    {
                                        "delta": {"content": "", "role": "assistant"},
                                        "finish_reason": "stop",
                                    }
                                ],
                                "usage": chunk["usage"],
                            }
                        )
                except Exception as e:
                    import warnings

                    warnings.warn(f"Failed to parse stream chunk: {str(e)}")
                    continue

            return result

    # Audio endpoints
    elif endpoint == "audio":
        try:
            if isinstance(response, dict):
                parsed = GroqAudioResponse(**response)
                return {"text": parsed.text, "metadata": parsed.x_groq}
        except Exception as e:
            import warnings

            warnings.warn(f"Failed to parse audio response: {str(e)}")
            return response

    # Default case
    return response



import inspect
from typing import Optional

from dotenv import load_dotenv

from lion.service import Service, register_service

from .api_endpoints.data_models import GroqAudioRequest, GroqChatCompletionRequest
from .GroqModel import GroqModel

load_dotenv()


@register_service
class GroqService(Service):
    def __init__(
        self,
        api_key: str,
        name: str | None = None,
    ):
        """Initialize the Groq service."""
        super().__setattr__("_initialized", False)
        self.api_key = api_key
        self.name = name
        self.rate_limiters = {}  # model: RateLimiter
        super().__setattr__("_initialized", True)

    def __setattr__(self, key, value):
        """Prevent modification of critical attributes after initialization."""
        if getattr(self, "_initialized", False) and key in ["api_key"]:
            raise AttributeError(
                f"Cannot modify '{key}' after initialization. "
                f"Please create a new service instance for different credentials."
            )
        super().__setattr__(key, value)

    def check_rate_limiter(
        self,
        groq_model: GroqModel,
        limit_requests: int | None = None,
        limit_tokens: int | None = None,
    ) -> GroqModel:
        """Check and update rate limiter for the model."""
        if groq_model.model not in self.rate_limiters:
            if groq_model.rate_limiter:
                self.rate_limiters[groq_model.model] = groq_model.rate_limiter
        else:
            groq_model.rate_limiter = self.rate_limiters[groq_model.model]
            if limit_requests is not None:
                groq_model.rate_limiter.limit_requests = limit_requests
            if limit_tokens is not None:
                groq_model.rate_limiter.limit_tokens = limit_tokens

        return groq_model

    @staticmethod
    def match_data_model(task_name: str) -> dict:
        """Match task name to appropriate request model mapping."""
        if task_name in ["chat/completions", "create_chat_completion"]:
            return {"request_body": GroqChatCompletionRequest}
        elif task_name in ["audio/transcriptions", "create_transcription"]:
            return {"request_body": GroqAudioRequest}
        elif task_name in ["audio/translations", "create_translation"]:
            return {"request_body": GroqAudioRequest}
        raise ValueError(f"Unknown task: {task_name}")

    @classmethod
    def list_tasks(cls) -> list[str]:
        """List available tasks."""
        methods = []
        for name, member in inspect.getmembers(cls, predicate=inspect.isfunction):
            if name not in [
                "__init__",
                "__setattr__",
                "check_rate_limiter",
                "match_data_model",
                "list_tasks",
            ]:
                methods.append(name)
        return methods

    # Chat Completions
    def create_chat_completion(
        self,
        model: str,
        limit_tokens: int | None = None,
        limit_requests: int | None = None,
    ) -> GroqModel:
        """Create a chat completion model."""
        model_obj = GroqModel(
            model=model,
            api_key=self.api_key,
            endpoint="chat/completions",
            method="POST",
            content_type="application/json",
        )

        return self.check_rate_limiter(
            model_obj,
            limit_requests=limit_requests,
            limit_tokens=limit_tokens,
        )

    # Audio
    def create_transcription(
        self,
        model: str,
        limit_requests: int | None = None,
    ) -> GroqModel:
        """Create an audio transcription model."""
        model_obj = GroqModel(
            model=model,
            api_key=self.api_key,
            endpoint="audio/transcriptions",
            method="POST",
        )

        return self.check_rate_limiter(
            model_obj,
            limit_requests=limit_requests,
        )

    def create_translation(
        self,
        model: str,
        limit_requests: int | None = None,
    ) -> GroqModel:
        """Create an audio translation model."""
        model_obj = GroqModel(
            model=model,
            api_key=self.api_key,
            endpoint="audio/translations",
            method="POST",
        )

        return self.check_rate_limiter(
            model_obj,
            limit_requests=limit_requests,
        )



from typing import Dict, List, Optional

from pydantic import Field, SerializeAsAny, model_validator

from ...data_models import OpenAIEndpointRequestBody
from .message_models import Message
from .response_format import ResponseFormat
from .stream_options import StreamOptions
from .tool_choice_models import ToolChoice as ToolChoiceObj
from .tool_models import Tool
from .types import ServiceTier
from .types import ToolChoice as ToolChoiceStr


class OpenAIChatCompletionRequestBody(OpenAIEndpointRequestBody):
    messages: SerializeAsAny[list[Message]] = Field(
        description="A list of messages comprising the conversation so far."
    )

    model: str = Field(description="ID of the model to use.")

    frequency_penalty: float | None = Field(
        0,
        ge=-2.0,
        le=2.0,
        description=(
            "Number between -2.0 and 2.0. Positive values penalize new tokens "
            "based on their existing frequency in the text so far, decreasing "
            "the model's likelihood to repeat the same line verbatim."
        ),
    )

    logit_bias: dict[str, float] | None = Field(
        None,
        description=(
            "Modify the likelihood of specified tokens appearing in the "
            "completion. Accepts a JSON object that maps tokens (specified by "
            "their token ID in the tokenizer) to an associated "
            "bias value from -100 to 100."
        ),
    )

    logprobs: bool | None = Field(
        False,
        description=(
            "Whether to return log probabilities of the output tokens or not. "
            "If true, returns the log probabilities of each output token "
            "returned in the content of message."
        ),
    )

    top_logprobs: int | None = Field(
        None,
        ge=0,
        le=20,
        description=(
            "An integer between 0 and 20 specifying the number of most likely "
            "tokens to return at each token position, each with an associated "
            "log probability. logprobs must be true if this parameter "
            "is used."
        ),
    )

    max_completion_tokens: int | None = Field(
        None,
        description=(
            "The maximum number of tokens that can be generated in the chat "
            "completion. The total length of input tokens and generated "
            "tokens is limited by the model's context length."
        ),
    )

    n: int | None = Field(
        1,
        ge=1,
        description=(
            "How many chat completion choices to generate for each input "
            "message. Note that you will be charged based on the number of "
            "generated tokens across all of the choices. Keep n as 1 to "
            "minimize costs."
        ),
    )

    presence_penalty: float | None = Field(
        0,
        ge=-2.0,
        le=2.0,
        description=(
            "Number between -2.0 and 2.0. Positive values penalize new tokens "
            "based on whether they appear in the text so far, increasing the "
            "model's likelihood to talk about new topics."
        ),
    )

    response_format: ResponseFormat | None = Field(
        None,
        description=(
            "An object specifying the format that the model must output. "
            "Compatible with GPT-4o, GPT-4o mini, GPT-4 Turbo and all GPT-3.5 "
            "Turbo models newer than gpt-3.5-turbo-1106. Setting to "
            "{ 'type': 'json_schema', 'json_schema': {...} } enables "
            "Structured Outputs. Setting to { 'type': 'json_object' } enables "
            "JSON mode. Important: when using JSON mode, you must also "
            "instruct the model to produce JSON via a system or user message."
        ),
    )

    seed: int | None = Field(
        None,
        description=(
            "This feature is in Beta. If specified, our system will make a "
            "best effort to sample deterministically, such that repeated "
            "requests with the same seed and parameters should return the "
            "same result."
        ),
    )

    service_tier: ServiceTier | None = Field(
        None,
        description=(
            "Specifies the latency tier to use for processing the request. "
            "This parameter is relevant for customers subscribed to the scale "
            "tier service."
        ),
    )

    stop: str | list[str] | None = Field(
        None,
        max_items=4,
        description=(
            "Up to 4 sequences where the API will stop generating further " "tokens."
        ),
    )

    stream: bool | None = Field(
        False,
        description=(
            "If set, partial message deltas will be sent, like in ChatGPT. "
            "Tokens will be sent as data-only server-sent events as they "
            "become available, with the stream terminated by a data: [DONE] "
            "message."
        ),
    )

    stream_options: StreamOptions | None = Field(
        None,
        description=(
            "Options for streaming response. Only set this when you set "
            "stream: true."
        ),
    )

    temperature: float | None = Field(
        1.0,
        ge=0,
        le=2,
        description=(
            "What sampling temperature to use, between 0 and 2. Higher values "
            "like 0.8 will make the output more random, while lower values "
            "like 0.2 will make it more focused and deterministic."
        ),
    )

    top_p: float | None = Field(
        1.0,
        ge=0,
        le=1,
        description=(
            "An alternative to sampling with temperature, called nucleus "
            "sampling, where the model considers the results of the tokens "
            "with top_p probability mass. So 0.1 means only the tokens "
            "comprising the top 10% probability mass are considered."
        ),
    )

    tools: list[Tool] | None = Field(
        None,
        max_items=128,
        description=(
            "A list of tools the model may call. Currently, only functions "
            "are supported as a tool. Use this to provide a list of functions "
            "the model may generate JSON inputs for."
        ),
    )

    tool_choice: ToolChoiceStr | ToolChoiceObj | None = Field(
        None,
        description=(
            "Controls which (if any) tool is called by the model. 'none' "
            "means the model will not call a tool and instead generates a "
            "message. 'auto' means the model can pick between generating a "
            "message or calling a tool. 'required' means the model must call "
            "a tool."
        ),
    )

    parallel_tool_calls: bool | None = Field(
        True,
        decription="Whether to enable parallel " "function calling during tool use.",
    )

    user: str | None = Field(
        None,
        description=(
            "A unique identifier representing your end-user, which can help "
            "OpenAI to monitor and detect abuse."
        ),
    )

    @model_validator(mode="after")
    def validate_request(self) -> "OpenAIChatCompletionRequestBody":
        if self.tools:
            self.model_fields["tool_choice"].default = "auto"
        else:
            self.model_fields["tool_choice"].default = "none"

        if self.top_logprobs is not None and not self.logprobs:
            raise ValueError("logprobs must be true when top_logprobs is set")
        return self



imported_models = {}


def match_response(request_model, response: dict | list):
    global imported_models

    endpoint = request_model.endpoint

    if endpoint == "generate":
        if isinstance(response, list):
            if "OllamaStreamCompletionResponseBody" not in imported_models:
                from .completion.response_body import OllamaStreamCompletionResponseBody

                imported_models["OllamaStreamCompletionResponseBody"] = (
                    OllamaStreamCompletionResponseBody
                )
            if "OllamaCompletionResponseBody" not in imported_models:
                from .completion.response_body import OllamaCompletionResponseBody

                imported_models["OllamaCompletionResponseBody"] = (
                    OllamaCompletionResponseBody
                )

            result = []
            for item in response[:-1]:
                result.append(
                    imported_models["OllamaStreamCompletionResponseBody"](**item)
                )

            result.append(
                imported_models["OllamaCompletionResponseBody"](**response[-1])
            )
            return result
        else:
            if "OllamaCompletionResponseBody" not in imported_models:
                from .completion.response_body import OllamaCompletionResponseBody

                imported_models["OllamaCompletionResponseBody"] = (
                    OllamaCompletionResponseBody
                )
            return imported_models["OllamaCompletionResponseBody"](**response)

    elif endpoint == "chat":
        if isinstance(response, list):
            if "OllamaStreamChatCompletionResponseBody" not in imported_models:
                from .chat_completion.response_body import (
                    OllamaStreamChatCompletionResponseBody,
                )

                imported_models["OllamaStreamChatCompletionResponseBody"] = (
                    OllamaStreamChatCompletionResponseBody
                )
            if "OllamaChatCompletionResponseBody" not in imported_models:
                from .chat_completion.response_body import (
                    OllamaChatCompletionResponseBody,
                )

                imported_models["OllamaChatCompletionResponseBody"] = (
                    OllamaChatCompletionResponseBody
                )

            result = []
            for item in response[:-1]:
                result.append(
                    imported_models["OllamaStreamChatCompletionResponseBody"](**item)
                )

            result.append(
                imported_models["OllamaChatCompletionResponseBody"](**response[-1])
            )
            return result
        else:
            if "OllamaChatCompletionResponseBody" not in imported_models:
                from .chat_completion.response_body import (
                    OllamaChatCompletionResponseBody,
                )

                imported_models["OllamaChatCompletionResponseBody"] = (
                    OllamaChatCompletionResponseBody
                )
            return imported_models["OllamaChatCompletionResponseBody"](**response)

    elif endpoint == "embed":
        if "OllamaEmbeddingResponseBody" not in imported_models:
            from .embedding.response_body import OllamaEmbeddingResponseBody

            imported_models["OllamaEmbeddingResponseBody"] = OllamaEmbeddingResponseBody
        return imported_models["OllamaEmbeddingResponseBody"](**response)

    elif endpoint == "create":
        if "OllamaCreateModelResponseBody" not in imported_models:
            from .model.create_model import OllamaCreateModelResponseBody

            imported_models["OllamaCreateModelResponseBody"] = (
                OllamaCreateModelResponseBody
            )
        if isinstance(response, list):
            return [
                imported_models["OllamaCreateModelResponseBody"](**res)
                for res in response
            ]
        else:  # dict
            return imported_models["OllamaCreateModelResponseBody"](**response)

    elif endpoint == "tags":
        if "OllamaListLocalModelsResponseBody" not in imported_models:
            from .model.list_model import OllamaListLocalModelsResponseBody

            imported_models["OllamaListLocalModelsResponseBody"] = (
                OllamaListLocalModelsResponseBody
            )
        return imported_models["OllamaListLocalModelsResponseBody"](**response)

    elif endpoint == "show":
        if "OllamaShowModelResponseBody" not in imported_models:
            from .model.show_model import OllamaShowModelResponseBody

            imported_models["OllamaShowModelResponseBody"] = OllamaShowModelResponseBody
        return imported_models["OllamaShowModelResponseBody"](**response)

    elif endpoint == "pull":
        if "OllamaPullModelResponseBody" not in imported_models:
            from .model.pull_model import OllamaPullModelResponseBody

            imported_models["OllamaPullModelResponseBody"] = OllamaPullModelResponseBody
        if isinstance(response, list):
            return [
                imported_models["OllamaPullModelResponseBody"](**res)
                for res in response
            ]
        else:  # dict
            return imported_models["OllamaPullModelResponseBody"](**response)

    elif endpoint == "push":
        if "OllamaPushModelResponseBody" not in imported_models:
            from .model.push_model import OllamaPushModelResponseBody

            imported_models["OllamaPushModelResponseBody"] = OllamaPushModelResponseBody
        if isinstance(response, list):
            return [
                imported_models["OllamaPushModelResponseBody"](**res)
                for res in response
            ]
        else:  # dict
            return imported_models["OllamaPushModelResponseBody"](**response)

    elif endpoint == "ps":
        if "OllamaListRunningModelsResponseBody" not in imported_models:
            from .model.list_model import OllamaListRunningModelsResponseBody

            imported_models["OllamaListRunningModelsResponseBody"] = (
                OllamaListRunningModelsResponseBody
            )
        return imported_models["OllamaListRunningModelsResponseBody"](**response)

    elif not response:
        return

    else:
        raise ValueError(
            "There is no standard response model for the provided request and response"
        )



from typing import List, Literal

from pydantic import Field

from ..data_models import OpenAIEndpointResponseBody


class OpenAIModelResponseBody(OpenAIEndpointResponseBody):
    id: str = Field(
        description="The model identifier, which can be referenced in the API endpoints.",
    )

    created: int = Field(
        description="The Unix timestamp (in seconds) when the model was created."
    )

    object: Literal["model"] = Field(
        description="The object type, which is always 'model'."
    )

    owned_by: str = Field(description="The organization that owns the model.")


class OpenAIListModelResponseBody(OpenAIEndpointResponseBody):
    object: Literal["list"] = Field(
        description="The object type, which is always 'list'."
    )

    data: list[OpenAIModelResponseBody] = Field(
        description="The list of model objects."
    )



from lion.core.models.schema_model import SchemaModel
from lion.libs.constants import UNDEFINED, UndefinedType


class TimedFuncCallConfig(SchemaModel):
    initial_delay: int = 0
    retry_default: str | UndefinedType = UNDEFINED
    retry_timeout: int | None = None
    retry_timing: bool = False
    error_msg: str | None = None
    error_map: dict | None = None


class RetryConfig(SchemaModel):
    num_retries: int = 0
    initial_delay: int = 0
    retry_delay: int = 0
    backoff_factor: int = 1
    retry_default: str | UndefinedType = UNDEFINED
    retry_timeout: int | None = None
    retry_timing: bool = False
    verbose_retry: bool = False
    error_msg: str | None = None
    error_map: dict | None = None



import os
import warnings
from datetime import datetime
from pathlib import Path

import yaml
from dotenv import load_dotenv
from pydantic import BaseModel, ConfigDict, Field, field_serializer, model_validator

from lion.integrations.perplexity_.api_endpoints.api_request import PerplexityRequest
from lion.integrations.perplexity_.api_endpoints.chat_completions.request.request_body import (
    PerplexityChatCompletionRequestBody,
)
from lion.integrations.perplexity_.api_endpoints.match_response import match_response
from lion.service.rate_limiter import RateLimiter, RateLimitError
from lion.service.service_util import invoke_retry
from lion.service.token_calculator import TiktokenCalculator

load_dotenv()
path = Path(__file__).parent

price_config_file_name = path / "perplexity_price_data.yaml"
max_output_token_file_name = path / "perplexity_max_output_token_data.yaml"


class PerplexityModel(BaseModel):
    model: str = Field(description="ID of the model to use.")

    request_model: PerplexityRequest = Field(description="Making requests")

    rate_limiter: RateLimiter = Field(description="Rate Limiter to track usage")

    text_token_calculator: TiktokenCalculator = Field(
        default=None, description="Token Calculator"
    )

    estimated_output_len: int = Field(
        default=0, description="Expected output len before making request"
    )

    model_config = ConfigDict(extra="forbid")

    @model_validator(mode="before")
    @classmethod
    def parse_input(cls, data: dict):
        if not isinstance(data, dict):
            raise ValueError("Invalid init param")

        # parse request model
        request_model_params = {
            "api_key": data.pop("api_key", None),
            "endpoint": data.pop("endpoint", None),
            "method": data.pop("method", None),
            "content_type": data.pop("content_type", None),
        }
        try:
            api_key = os.getenv(request_model_params["api_key"], None)
            if api_key:
                request_model_params["api_key"] = api_key
        except Exception:
            pass

        data["request_model"] = PerplexityRequest(**request_model_params)

        # parse rate limiter
        if "rate_limiter" not in data:
            rate_limiter_params = {}
            if limit_tokens := data.pop("limit_tokens", None):
                rate_limiter_params["limit_tokens"] = limit_tokens
            if limit_requests := data.pop("limit_requests", None):
                rate_limiter_params["limit_requests"] = limit_requests

            data["rate_limiter"] = RateLimiter(**rate_limiter_params)

        # parse token calculator
        try:
            # Perplexity uses cl100k_base encoding like OpenAI
            text_calc = TiktokenCalculator(encoding_name="cl100k_base")
            data["text_token_calculator"] = text_calc
        except Exception:
            pass

        return data

    @field_serializer("request_model")
    def serialize_request_model(self, value: PerplexityRequest):
        return value.model_dump(exclude_unset=True)

    @invoke_retry(max_retries=3, base_delay=1, max_delay=60)
    async def invoke(
        self,
        request_body: PerplexityChatCompletionRequestBody,
        estimated_output_len: int = 0,
        output_file=None,
        parse_response=True,
    ):
        if request_model := getattr(request_body, "model"):
            if request_model != self.model:
                raise ValueError(
                    f"Request model does not match. Model is {self.model}, but request is made for {request_model}."
                )

        # check remaining rate limit
        input_token_len = await self.get_input_token_len(request_body)

        if getattr(request_body, "max_tokens", None):
            estimated_output_len = request_body.max_tokens

        invoke_viability_result = self.verify_invoke_viability(
            input_tokens_len=input_token_len, estimated_output_len=estimated_output_len
        )
        if not invoke_viability_result:
            raise RateLimitError(
                message="Rate limit reached for requests",
                input_token_len=input_token_len,
                estimated_output_len=estimated_output_len,
            )

        try:
            if getattr(request_body, "stream", None):
                return await self.stream(
                    request_body, output_file=output_file, parse_response=parse_response
                )

            response_body, response_headers = await self.request_model.invoke(
                json_data=request_body,
                output_file=output_file,
                with_response_header=True,
                parse_response=False,
            )

            if response_body:
                # Update rate limit based on usage
                if response_body.get("usage"):
                    total_token_usage = response_body["usage"]["total_tokens"]
                    if date_str := response_headers.get("date"):
                        self.rate_limiter.update_rate_limit(date_str, total_token_usage)
                    else:
                        self.rate_limiter.update_rate_limit(None, total_token_usage)
                else:
                    self.rate_limiter.update_rate_limit(None)

            if parse_response:
                return match_response(self.request_model, response_body)
            else:
                return response_body

        except Exception as e:
            raise e

    async def stream(
        self,
        request_body: PerplexityChatCompletionRequestBody,
        output_file=None,
        parse_response=True,
        verbose=True,
    ):
        response_chunks = []
        response_headers = None

        async for chunk in self.request_model.stream(
            json_data=request_body,
            output_file=output_file,
            with_response_header=True,
            verbose=verbose,
        ):
            if isinstance(chunk, dict):
                if "headers" in chunk:
                    response_headers = chunk["headers"]
                else:
                    response_chunks.append(chunk)

        # Update rate limit if we have usage information
        if response_chunks and response_chunks[-1].get("usage"):
            total_token_usage = response_chunks[-1]["usage"]["total_tokens"]
            if date_str := response_headers.get("date"):
                self.rate_limiter.update_rate_limit(date_str, total_token_usage)
            else:
                self.rate_limiter.update_rate_limit(None, total_token_usage)

        if parse_response:
            return match_response(self.request_model, response_chunks)
        else:
            return response_chunks

    async def get_input_token_len(
        self, request_body: PerplexityChatCompletionRequestBody
    ):
        if request_model := getattr(request_body, "model"):
            if request_model != self.model:
                raise ValueError(
                    f"Request model does not match. Model is {self.model}, but request is made for {request_model}."
                )

        total_tokens = 0
        for message in request_body.messages:
            total_tokens += self.text_token_calculator.calculate(message.content)

        return total_tokens

    def verify_invoke_viability(
        self, input_tokens_len: int = 0, estimated_output_len: int = 0
    ):
        self.rate_limiter.release_tokens()

        estimated_output_len = (
            estimated_output_len
            if estimated_output_len != 0
            else self.estimated_output_len
        )
        if estimated_output_len == 0:
            with open(max_output_token_file_name) as file:
                output_token_config = yaml.safe_load(file)
                estimated_output_len = output_token_config.get(self.model, 0)
                self.estimated_output_len = estimated_output_len

        if self.rate_limiter.check_availability(input_tokens_len, estimated_output_len):
            return True
        else:
            return False

    def estimate_text_price(
        self,
        input_text: str,
        estimated_num_of_output_tokens: int = 0,
    ):
        if self.text_token_calculator is None:
            raise ValueError("Token calculator not available")

        num_of_input_tokens = self.text_token_calculator.calculate(input_text)

        with open(price_config_file_name) as file:
            price_config = yaml.safe_load(file)

        model_price_info_dict = price_config["model"][self.model]
        estimated_price = (
            model_price_info_dict["input_tokens"] * num_of_input_tokens
            + model_price_info_dict["output_tokens"] * estimated_num_of_output_tokens
        )

        return estimated_price



from .extend_df import extend_dataframe
from .read import read_csv, read_json
from .remove_rows import remove_rows
from .replace_keywords import replace_keywords
from .save import to_csv, to_excel
from .search_keywords import search_dataframe_keywords
from .to_df import to_df
from .update_cells import update_cells

__all__ = [
    "to_df",
    "replace_keywords",
    "search_dataframe_keywords",
    "read_csv",
    "read_json",
    "to_csv",
    "to_excel",
    "extend_dataframe",
    "remove_rows",
    "update_cells",
]



import logging
from typing import Any, Protocol, TypeVar, runtime_checkable

from typing_extensions import get_protocol_members

T = TypeVar("T")


@runtime_checkable
class Adapter(Protocol):

    obj_key: str

    @classmethod
    def from_obj(
        cls, subj_cls: type[T], obj: Any, /, **kwargs
    ) -> dict | list[dict]: ...

    @classmethod
    def to_obj(cls, subj: T, /, **kwargs) -> Any: ...


adapter_members = get_protocol_members(Adapter)  # duck typing


class AdapterRegistry:

    _adapters: dict[str, Adapter] = {}

    @classmethod
    def list_adapters(cls) -> list[tuple[str | type, ...]]:
        return list(cls._adapters.keys())

    @classmethod
    def register(cls, adapter: type[Adapter]) -> None:
        for member in adapter_members:
            if not hasattr(adapter, member):
                _str = getattr(adapter, "obj_key", None) or repr(adapter)
                _str = _str[:50] if len(_str) > 50 else _str
                raise AttributeError(f"Adapter {_str} missing required methods.")

        if isinstance(adapter, type):
            cls._adapters[adapter.obj_key] = adapter()
        else:
            cls._adapters[adapter.obj_key] = adapter

    @classmethod
    def get(cls, obj_key: type | str) -> Adapter:
        try:
            return cls._adapters[obj_key]
        except Exception as e:
            logging.error(f"Error getting adapter for {obj_key}. Error: {e}")

    @classmethod
    def adapt_from(
        cls, subj_cls: type[T], obj: Any, obj_key: type | str, **kwargs
    ) -> dict | list[dict]:
        try:
            return cls.get(obj_key).from_obj(subj_cls, obj, **kwargs)
        except Exception as e:
            logging.error(f"Error adapting data from {obj_key}. Error: {e}")
            raise e

    @classmethod
    def adapt_to(cls, subj: T, obj_key: type | str, **kwargs) -> Any:
        try:
            return cls.get(obj_key).to_obj(subj, **kwargs)
        except Exception as e:
            logging.error(f"Error adapting data to {obj_key}. Error: {e}")
            raise e


__all__ = ["AdapterRegistry", "Adapter"]



from datetime import timezone
from enum import Enum

from lion.protocols.configs import (
    BranchConfig,
    LionIDConfig,
    MessageConfig,
    RetryConfig,
    TimedFuncCallConfig,
    iModelConfig,
)
from lion.protocols.configs.log_config import LogConfig


class BaseSystemFields(str, Enum):
    ln_id = "ln_id"
    TIMESTAMP = "timestamp"
    METADATA = "metadata"
    EXTRA_FIELDS = "extra_fields"
    CONTENT = "content"
    CREATED = "created"
    EMBEDDING = "embedding"


DEFAULT_ID_CONFIG = LionIDConfig(
    n=36,
    random_hyphen=True,
    num_hyphens=4,
    hyphen_start_index=6,
    hyphen_end_index=-6,
    prefix="ao",
    postfix="",
)

DEFAULT_TIMED_FUNC_CALL_CONFIG = TimedFuncCallConfig(
    initial_delay=0,
    retry_timeout=None,
    retry_timing=False,
    error_msg=None,
    error_map=None,
)

DEFAULT_RETRY_CONFIG = RetryConfig(
    num_retries=0,
    initial_delay=0,
    retry_delay=0,
    backoff_factor=1,
    retry_timeout=None,
    retry_timing=False,
    verbose_retry=False,
    error_msg=None,
    error_map=None,
)


DEFAULT_CHAT_CONFIG = iModelConfig(
    provider="openai",
    task="chat",
    model="gpt-4o",
    api_key="OPENAI_API_KEY",
)


DEFAULT_RETRY_iMODEL_CONFIG = iModelConfig(
    provider="openai",
    task="chat",
    model="gpt-4o-mini",
    api_key="OPENAI_API_KEY",
)


DEFAULT_MESSAGE_CONFIG = MessageConfig(
    validation_mode="raise",
    auto_retries=False,
    max_retries=0,
    allow_actions=True,
    auto_invoke_action=True,
)


DEFAULT_MESSAGE_LOG_CONFIG = LogConfig(
    persist_dir="./data/logs",
    subfolder="messages",
    file_prefix="message_",
    capacity=128,
    auto_save_on_exit=True,
    clear_after_dump=True,
    use_timestamp=True,
    extension=".csv",
)

DEFAULT_ACTION_LOG_CONFIG = LogConfig(
    persist_dir="./data/logs",
    subfolder="actions",
    file_prefix="action_",
    capacity=128,
    auto_save_on_exit=True,
    clear_after_dump=True,
    use_timestamp=True,
    extension=".csv",
)

DEFAULT_BRANCH_CONFIG = BranchConfig(
    name=None,
    user=None,
    message_log_config=DEFAULT_MESSAGE_LOG_CONFIG,
    action_log_config=DEFAULT_ACTION_LOG_CONFIG,
    message_config=DEFAULT_MESSAGE_CONFIG,
    auto_register_tools=True,
    action_call_config=DEFAULT_TIMED_FUNC_CALL_CONFIG,
    imodel_config=DEFAULT_CHAT_CONFIG,
    retry_imodel_config=DEFAULT_RETRY_iMODEL_CONFIG,
)


DEFAULT_TIMEZONE = timezone.utc
BASE_LION_FIELDS = set(BaseSystemFields.__members__.values())


class Settings:

    class Config:
        ID: LionIDConfig = DEFAULT_ID_CONFIG
        RETRY: RetryConfig = DEFAULT_RETRY_CONFIG
        TIMED_CALL: TimedFuncCallConfig = DEFAULT_TIMED_FUNC_CALL_CONFIG
        TIMEZONE: timezone = DEFAULT_TIMEZONE

    class Branch:
        BRANCH: BranchConfig = DEFAULT_BRANCH_CONFIG

    class iModel:
        CHAT: dict = {"model": "openai/gpt-4o"}
        PARSE: iModelConfig = DEFAULT_CHAT_CONFIG


# File: autoos/setting.py



from typing import List, Literal, Optional, TypeAlias

from pydantic import BaseModel, ConfigDict, Field, model_validator

from .types import Detail


class TextContentPart(BaseModel):
    type: Literal["text"] = Field(description="The type of the content part.")
    text: str = Field(description="The text content.")


class ImageURL(BaseModel):
    url: str = Field(description="The URL of the image.")
    detail: Detail = Field("auto", description="The detail level of the image.")
    model_config = ConfigDict(use_enum_values=True)


class ImageContentPart(BaseModel):
    type: Literal["image_url"] = Field(description="The type of the content part.")
    image_url: ImageURL = Field(description="The image URL and detail level.")


ContentPart: TypeAlias = TextContentPart | ImageContentPart


class Function(BaseModel):
    name: str = Field(description="The name of the function to call.")
    arguments: str = Field(
        description="The arguments to call the function with, as JSON."
    )


class ToolCall(BaseModel):
    id: str = Field(description="The ID of the tool call.")
    type: Literal["function"] = Field(
        description="The type of the tool. Only 'function' is supported."
    )
    function: Function = Field(description="The function that the model called.")


class SystemMessage(BaseModel):
    role: Literal["system"]
    content: str = Field(description="The content of the system message.")

    name: str | None = Field(
        None,
        description="An optional name for the participant."
        " Provides the model information "
        "to differentiate between participants of the same role.",
    )


class UserMessage(BaseModel):
    role: Literal["user"]
    content: str | list[ContentPart] = Field(
        description="The content of the user message"
    )

    name: str | None = Field(
        None,
        description="An optional name for the participant."
        " Provides the model information "
        "to differentiate between participants of the same role.",
    )


class AssistantMessage(BaseModel):
    role: Literal["assistant"]
    content: str | None = Field(
        None,
        description="The contents of the assistant message."
        " Required unless tool_calls is specified.",
    )

    refusal: str | None = Field(
        None, description="The refusal message by the assistant."
    )

    name: str | None = Field(
        None,
        description="An optional name for the participant."
        " Provides the model information "
        "to differentiate between participants of the same role.",
    )

    tool_calls: list[ToolCall] | None = Field(
        None,
        description="The tool calls generated by the model," " such as function calls.",
    )

    @model_validator(mode="after")
    def validate_content(self):
        if not self.content and not self.tool_calls:
            raise ValueError("Assistant messages require content or tool_calls")
        return self


class ToolMessage(BaseModel):
    role: Literal["tool"]
    content: str | list = Field(description="The contents of the tool message.")

    tool_calls_id: str = Field(
        description="Tool call that this message is responding to."
    )


Message: TypeAlias = SystemMessage | UserMessage | AssistantMessage | ToolMessage



from .image_edit_models import OpenAIImageEditRequestBody
from .image_models import OpenAIImageRequestBody
from .image_variation_models import OpenAIImageVariationRequestBody

__all__ = [
    "OpenAIImageRequestBody",
    "OpenAIImageEditRequestBody",
    "OpenAIImageVariationRequestBody",
]



from pydantic import Field

from ..data_models import OllamaEndpointResponseBody


class OllamaEmbeddingResponseBody(OllamaEndpointResponseBody):
    model: str = Field(None, description="The model name")

    embeddings: list = Field(
        None, description="The generated embeddings for the text or list of text"
    )

    total_duration: int = Field(None, description="Time spent generating the response")

    load_duration: int = Field(
        None, description="Time spent in nanoseconds loading the model"
    )

    prompt_eval_count: int = Field(None, description="Number of tokens in the prompt")



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from lion.core.typing import ID, Any, BaseModel, JsonValue, Note, override
from lion.libs.parse import to_str
from lion.libs.utils import copy, is_same_dtype

from .message import MessageFlag, MessageRole, RoledMessage


def prepare_assistant_response(
    assistant_response: BaseModel | list[BaseModel] | dict | str | Any, /
) -> Note:
    """
    Prepare an assistant's response for storage and transmission.

    This function handles various input formats including:
    - Single model outputs (response.choices[0].message.content)
    - Streaming responses (response[i].choices[0].delta.content)
    - Direct content in dictionaries or strings

    Args:
        assistant_response: The response content in any supported format

    Returns:
        Note: Formatted response content
    """
    if assistant_response:
        content = Note()
        # Handle model.choices[0].message.content format
        if isinstance(assistant_response, BaseModel):
            content["assistant_response"] = (
                assistant_response.choices[0].message.content or ""
            )
            content["model_response"] = assistant_response.model_dump(
                exclude_none=True, exclude_unset=True
            )
        # Handle streaming response[i].choices[0].delta.content format
        elif isinstance(assistant_response, list):
            if is_same_dtype(assistant_response, BaseModel):
                msg = "".join(
                    [i.choices[0].delta.content or "" for i in assistant_response]
                )
                content["assistant_response"] = msg
                content["model_response"] = [
                    i.model_dump(
                        exclude_none=True,
                        exclude_unset=True,
                    )
                    for i in assistant_response
                ]
            elif is_same_dtype(assistant_response, dict):
                msg = "".join(
                    [
                        i["choices"][0]["delta"]["content"] or ""
                        for i in assistant_response
                    ]
                )
                content["assistant_response"] = msg
                content["model_response"] = assistant_response
        elif isinstance(assistant_response, dict):
            if "content" in assistant_response:
                content["assistant_response"] = assistant_response["content"]
            elif "choices" in assistant_response:
                content["assistant_response"] = assistant_response["choices"][0][
                    "message"
                ]["content"]
            content["model_response"] = assistant_response
        elif isinstance(assistant_response, str):
            content["assistant_response"] = assistant_response
        else:
            content["assistant_response"] = to_str(assistant_response)
        return content
    else:
        return Note(assistant_response="")


class AssistantResponse(RoledMessage):
    """
    Represents a response from an assistant in the system.

    This class handles responses from AI assistants, supporting both direct
    text responses and structured model outputs. It provides access to both
    the formatted response content and the underlying model data.

    Example:
        >>> response = AssistantResponse(
        ...     assistant_response="The answer is 42",
        ...     sender="assistant",
        ...     recipient="user"
        ... )
        >>> print(response.response)
        'The answer is 42'
    """

    @override
    def __init__(
        self,
        assistant_response: BaseModel | JsonValue | MessageFlag,
        sender: ID.Ref | MessageFlag,
        recipient: ID.Ref | MessageFlag,
        protected_init_params: dict | None = None,
    ) -> None:
        """
        Initialize an AssistantResponse instance.

        Args:
            assistant_response: The content of the assistant's response
            sender: The sender of the response
            recipient: The recipient of the response
            protected_init_params: Protected initialization parameters
        """
        message_flags = [assistant_response, sender, recipient]

        if all(x == MessageFlag.MESSAGE_LOAD for x in message_flags):
            protected_init_params = protected_init_params or {}
            super().__init__(**protected_init_params)
            return

        if all(x == MessageFlag.MESSAGE_CLONE for x in message_flags):
            super().__init__(role=MessageRole.ASSISTANT)
            return

        super().__init__(
            role=MessageRole.ASSISTANT,
            sender=sender or "N/A",
            recipient=recipient,
        )

        content = prepare_assistant_response(assistant_response)
        if "model_response" in content:
            self.metadata["model_response"] = content.pop("model_response")
        self.content = content

    @property
    def response(self) -> str:
        """
        Get the assistant response content.

        Returns:
            str: The formatted content of the assistant's response
        """
        return copy(self.content["assistant_response"])

    @property
    def model_response(self) -> dict | list[dict]:
        """
        Get the underlying model response data.

        Returns:
            Union[dict, List[dict]]: The complete model response data
        """
        return copy(self.metadata.get("model_response", {}))

    @override
    def _format_content(self) -> dict[str, str]:
        return {"role": self.role.value, "content": self.response}



from pydantic import Field

from ..data_models import OpenAIEndpointPathParam


class OpenAIRetrieveFilePathParam(OpenAIEndpointPathParam):
    file_id: str = Field(description="The ID of the file to use for this request.")



import warnings
from typing import Any, Dict, Literal, Optional

from pydantic import BaseModel, Field, model_validator

# Suppress the specific warning about field name shadowing
warnings.filterwarnings(
    "ignore",
    message='Field name "schema" in "JSONSchema" '
    'shadows an attribute in parent "BaseModel"',
)


class JSONSchema(BaseModel):
    description: str | None = Field(
        None,
        description=(
            "A description of what the response format is for, used by the "
            "model to determine how to respond in the format."
        ),
    )

    name: str = Field(
        max_length=64,
        pattern="^[a-zA-Z0-9_-]+$",
        description=(
            "The name of the response format. Must be a-z, A-Z, 0-9, or "
            "contain underscores and dashes, with a maximum length of 64."
        ),
    )

    schema: dict[str, Any] | None = Field(
        None,
        description="The schema for the response "
        "format, described as a JSON Schema object.",
    )

    strict: bool | None = Field(
        False,
        description=(
            "Whether to enable strict schema adherence when generating the "
            "output. If set to true, the model will always follow the exact "
            "schema defined in the `schema` field. Only a subset of JSON "
            "Schema is supported when `strict` is `true`."
        ),
    )


class ResponseFormat(BaseModel):
    type: Literal["text", "json_object", "json_schema"] = Field(
        description="The type of response format being defined."
    )
    json_schema: JSONSchema | None = Field(
        None,
        description=(
            "The JSON schema to use when type is 'json_schema'. Required "
            "when type is 'json_schema'."
        ),
    )

    @model_validator(mode="after")
    def validate_response_format(self) -> "ResponseFormat":
        if self.type == "json_schema" and not self.json_schema:
            raise ValueError("json_schema is required when type is 'json_schema'")
        if self.type != "json_schema" and self.json_schema:
            raise ValueError(
                "json_schema should only be set when type is 'json_schema'"
            )
        return self



import base64
import math
import os
from io import BytesIO
from typing import Literal

import aiohttp
import yaml
from PIL import Image
from pydantic import BaseModel, Field, field_validator

image_token_config_file_name = os.path.join(
    os.path.dirname(__file__), "openai_image_token_data.yaml"
)


class OpenAIImageTokenCalculator(BaseModel):
    model: str = Field(description="ID of the model to use.")

    @field_validator("model")
    @classmethod
    def validate_model_image_function(cls, value: str):
        with open(image_token_config_file_name) as file:
            token_config = yaml.safe_load(file)

        token_config = token_config.get(value, None)

        if token_config is None:
            raise ValueError("The model does not have vision capabilities.")
        return value

    @staticmethod
    async def get_image_size(url: str):
        if url.startswith("data:image/jpeg;base64,"):
            # base64 encoded format
            image_base64 = url.split("data:image/jpeg;base64,")[1]
            image_base64.strip("{}")
            image_data = base64.b64decode(image_base64)
            image = Image.open(BytesIO(image_data))
        else:
            # image url
            async with aiohttp.ClientSession() as client:
                async with client.get(url=url) as response:
                    response.raise_for_status()

                    content = await response.read()

                    image = Image.open(BytesIO(content))

        return image.size

    async def calculate(
        self, url: str, detail: Literal["low", "high", "auto"] = "auto"
    ):
        if detail not in ["low", "high", "auto"]:
            raise ValueError(
                "Invalid detail option. Valid options are: low, high, auto"
            )

        with open(image_token_config_file_name) as file:
            token_config = yaml.safe_load(file)

        token_config = token_config.get(self.model, None)

        if token_config is None:
            raise ValueError("The model does not have vision capabilities.")

        width, height = await self.get_image_size(url)

        if detail == "low":
            return token_config["base"]
        if detail == "auto":
            # TODO: check "auto" option condition
            if width <= 512 and height <= 512:
                return token_config["base"]

        # Scale to fit within a 2048 x 2048 square
        if max(width, height) > 2048:
            scale_factor = 2048 / max(width, height)
            width = int(width * scale_factor)
            height = int(height * scale_factor)

        # Scale such that the shortest side is 768px
        if min(width, height) > 768:
            scale_factor = 768 / min(width, height)
            width = int(width * scale_factor)
            height = int(height * scale_factor)

        # Calculate the number of 512px squares
        num_squares = math.ceil(width / 512) * math.ceil(height / 512)

        return num_squares * token_config["tile"] + token_config["base"]



import inspect
import re
import warnings
from os import getenv

from .service import ServiceSetting

imported_services = {}


def match_service(service_name: str, **kwargs):
    service_name = service_name.lower()

    def check_key_match(key1, key2):
        if key1 is not None:
            key1 = getenv(key1) if getenv(key1) else key1
        if key2 is not None:
            key2 = getenv(key2) if getenv(key2) else key2
        return key1 == key2

    if service_name == "openai":
        from lion.integrations.openai_ import OpenAIService

        api_key = kwargs.get("api_key")
        openai_organization = kwargs.get("openai_organization")
        openai_project = kwargs.get("openai_project")

        for service in ServiceSetting.services.values():
            if isinstance(service, OpenAIService):
                if (
                    check_key_match(service.api_key, api_key)
                    and check_key_match(
                        service.openai_organization, openai_organization
                    )
                    and check_key_match(service.openai_project, openai_project)
                ):
                    return service

        # when no matching
        return OpenAIService(
            api_key=api_key,
            openai_organization=openai_organization,
            openai_project=openai_project,
        )

    elif service_name == "anthropic":
        from lion.integrations.anthropic_ import AnthropicService

        api_key = kwargs.get("api_key")
        api_version = kwargs.get("api_version", "2023-06-01")

        for service in ServiceSetting.services.values():
            if isinstance(service, AnthropicService):
                if check_key_match(service.api_key, api_key) and check_key_match(
                    service.api_version, api_version
                ):
                    return service

        # when no matching
        return AnthropicService(
            api_key=api_key,
            api_version=api_version,
        )

    elif service_name == "perplexity":
        from lion.integrations.perplexity_ import PerplexityService

        api_key = kwargs.get("api_key")
        api_version = kwargs.get("api_version", "2023-06-01")

        for service in ServiceSetting.services.values():
            if isinstance(service, PerplexityService):
                if check_key_match(service.api_key, api_key):
                    return service

        # when no matching
        return PerplexityService(api_key=api_key)

    elif service_name == "groq":
        from lion.integrations.groq_ import GroqService

        api_key = kwargs.get("api_key")
        for service in ServiceSetting.services.values():
            if isinstance(service, GroqService):
                if check_key_match(service.api_key, api_key):
                    return service

        # when no matching
        return GroqService(api_key=api_key)

    raise ValueError(f"Service {service_name} is not available")


def match_task_method(task_name, service_obj):
    def clean_name(name):
        return re.sub(r"[^a-zA-Z0-9]", "", name).lower()

    task_name = clean_name(task_name)
    if task_name.endswith("s"):
        task_name = task_name[:-1]
    methods = [
        method_name
        for method_name in dir(service_obj)
        if task_name in clean_name(method_name)
    ]
    return methods


def match_parameters(method, model, interval_tokens, interval_requests):
    method_params = inspect.signature(method).parameters
    kwargs = {}
    if "model" in method_params:
        kwargs["model"] = model
    if "limit_requests" in method_params:
        kwargs["limit_requests"] = interval_requests
    else:
        # no limit_requests but interval_requests provided
        if interval_requests is not None:
            warnings.warn(
                f"The system does not support tracking limit requests per minute for {method.__name__}"
            )
    if "limit_tokens" in method_params:
        kwargs["limit_tokens"] = interval_tokens
    else:
        # no limit_tokens but interval_tokens provided
        if interval_tokens is not None:
            warnings.warn(
                f"The system does not support tracking limit tokens per minute for {method.__name__}"
            )

    return kwargs



import pandas as pd

from .to_df import to_df


def replace_keywords(
    df: pd.DataFrame,
    keyword: str | dict[str, str],
    replacement: str | None = None,
    *,
    column: str | list[str] = "content",
    case_sensitive: bool = False,
    regex: bool = False,
    inplace: bool = False,
) -> pd.DataFrame | None:
    """
    Replaces occurrences of specified keyword(s) with replacement string(s) in DataFrame column(s).

    Args:
        df: The DataFrame to modify.
        keyword: The keyword to be replaced. Can be a string or a dictionary of {old: new} pairs.
        replacement: The string to replace the keyword with. Ignored if keyword is a dict.
        column: The column(s) in which to perform the replacement. Can be a string or list of strings.
        case_sensitive: If True, performs a case-sensitive replacement.
        regex: If True, treat keyword(s) as regular expressions.
        inplace: If True, modifies the DataFrame in place and returns None.

    Returns:
        Modified DataFrame if inplace is False, None otherwise.

    Example:
        >>> df = pd.DataFrame({'content': ['apple pie', 'banana split', 'cherry pie']})
        >>> replace_keyword(df, {'pie': 'tart', 'split': 'smoothie'}, column='content')
           content
        0  apple tart
        1  banana smoothie
        2  cherry tart
    """
    df_ = df if inplace else df.copy()

    if isinstance(column, str):
        column = [column]

    if isinstance(keyword, dict):
        for col in column:
            df_[col] = df_[col].replace(
                to_replace=keyword,
                regex=regex,
                case=case_sensitive,
            )
    else:
        for col in column:
            df_[col] = df_[col].replace(
                to_replace=keyword,
                value=replacement,
                regex=regex,
                case=case_sensitive,
            )

    if inplace:
        return None
    return to_df(df_)



from typing import List, Optional

from pydantic import BaseModel, Field, SerializeAsAny

from ..chat_completions.request.request_body import Message, Tool


class OpenAIChatModelTrainingFormat(BaseModel):
    messages: SerializeAsAny[list[Message]] = Field(
        description="A list of messages comprising the conversation so far."
    )

    tools: list[Tool] | None = Field(
        None, description="A list of tools may generate JSON inputs for."
    )

    parallel_tool_calls: bool | None = Field(
        True,
        decription="Whether to enable parallel function calling during tool use.",
    )


class OpenAICompletionsModelTrainingFormat(BaseModel):
    prompt: str = Field(description="The input prompt for this training example.")

    completion: str = Field(
        description="The desired completion for this training example."
    )



from lion.libs.parse import flatten, nget, ninsert, npop, nset, to_list
from lion.libs.utils import copy

from ..typing.pydantic_ import ConfigDict, Field, field_serializer
from ..typing.typing_ import UNDEFINED, Any, ItemsView, Iterator, ValuesView, override
from .base import BaseAutoModel

INDICE_TYPE = str | list[str | int]


class Note(BaseAutoModel):
    """Container for managing nested dictionary data structures.

    Provides:
    - Deep nested data access
    - Dictionary-like interface
    - Flattening capabilities
    - Update operations

    Example:
        ```python
        note = Note(
            user={
                "name": "John",
                "settings": {
                    "theme": "dark"
                }
            }
        )

        # Access nested data
        name = note.get(["user", "name"])
        theme = note["user"]["settings"]["theme"]

        # Update nested structure
        note.update(["user", "settings"], {"language": "en"})
        ```

    Attributes:
        content: Nested dictionary structure
        model_config: Configuration allowing arbitrary types
    """

    content: dict[str, Any] = Field(default_factory=dict)  # Nested data structure

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        use_enum_values=True,
        populate_by_name=True,
    )

    def __init__(self, **kwargs: Any) -> None:
        """Initialize Note with dictionary data.

        Args:
            **kwargs: Key-value pairs for initial content
        """
        super().__init__()
        self.content = kwargs

    @field_serializer("content")
    def _serialize_content(self, value: Any) -> dict[str, Any]:
        """Serialize content to dictionary format.

        Args:
            value: Content to serialize

        Returns:
            Deep copy of content dictionary
        """
        output_dict = copy(value, deep=True)
        return output_dict

    def to_dict(self) -> dict[str, Any]:
        """Convert Note to dictionary, excluding undefined values.

        Returns:
            Dictionary representation with UNDEFINED values removed
        """
        out = copy(self.content)
        for k, v in self.content.items():
            if v is UNDEFINED:
                out.pop(k)
        return out

    def pop(
        self,
        indices: INDICE_TYPE,
        /,
        default: Any = UNDEFINED,
    ) -> Any:
        """Remove and return item from nested structure.

        Args:
            indices: Path to item
            default: Value to return if not found

        Returns:
            Removed value or default

        Raises:
            KeyError: If path not found and no default
        """
        indices = to_list(indices, flatten=True, dropna=True)
        return npop(self.content, indices, default)

    def insert(self, indices: INDICE_TYPE, value: Any, /) -> None:
        """Insert value into nested structure at specified indices.

        Args:
            indices: Path where to insert
            value: Value to insert
        """
        indices = to_list(indices, flatten=True, dropna=True)
        ninsert(self.content, indices, value)

    def set(self, indices: INDICE_TYPE, value: Any, /) -> None:
        """Set value in nested structure at specified indices.

        Args:
            indices: Path where to set
            value: Value to set
        """
        indices = to_list(indices, flatten=True, dropna=True)
        if self.get(indices, None) is None:
            self.insert(indices, value)
        else:
            nset(self.content, indices, value)

    def get(
        self,
        indices: INDICE_TYPE,
        /,
        default: Any = UNDEFINED,
    ) -> Any:
        """Get value from nested structure at specified indices.

        Args:
            indices: Path to value
            default: Value to return if not found

        Returns:
            Value at path or default

        Raises:
            KeyError: If path not found and no default
        """
        indices = to_list(indices, flatten=True, dropna=True)
        return nget(self.content, indices, default)

    def keys(self, /, flat: bool = False, **kwargs: Any) -> list:
        """Get keys of the Note.

        Args:
            flat: If True, return flattened keys
            kwargs: Additional flattening options

        Returns:
            List of keys, optionally flattened
        """
        if flat:
            kwargs["coerce_keys"] = kwargs.get("coerce_keys", False)
            kwargs["coerce_sequence"] = kwargs.get("coerce_sequence", "list")
            return flatten(self.content, **kwargs).keys()
        return list(self.content.keys())

    def values(self, /, flat: bool = False, **kwargs: Any) -> ValuesView:
        """Get values of the Note.

        Args:
            flat: If True, return flattened values
            kwargs: Additional flattening options

        Returns:
            View of values, optionally flattened
        """
        if flat:
            kwargs["coerce_keys"] = kwargs.get("coerce_keys", False)
            kwargs["coerce_sequence"] = kwargs.get("coerce_sequence", "list")
            return flatten(self.content, **kwargs).values()
        return self.content.values()

    def items(self, /, flat: bool = False, **kwargs: Any) -> ItemsView:
        """Get items of the Note.

        Args:
            flat: If True, return flattened items
            kwargs: Additional flattening options

        Returns:
            View of items, optionally flattened
        """
        if flat:
            kwargs["coerce_keys"] = kwargs.get("coerce_keys", False)
            kwargs["coerce_sequence"] = kwargs.get("coerce_sequence", "list")
            return flatten(self.content, **kwargs).items()
        return self.content.items()

    def clear(self) -> None:
        """Clear all content."""
        self.content.clear()

    def update(
        self,
        indices: INDICE_TYPE,
        value: Any,
    ) -> None:
        """Update nested structure at specified indices.

        Args:
            indices: Location to update
            value: New value to set

        Raises:
            ValueError: If trying to update dict with non-dict
        """
        existing = None
        if not indices:
            existing = self.content
        else:
            existing = self.get(indices, None)

        if existing is None:
            if not isinstance(value, (list, dict)):
                value = [value]
            self.set(indices, value)

        if isinstance(existing, list):
            if isinstance(value, list):
                existing.extend(value)
            else:
                existing.append(value)

        elif isinstance(existing, dict):
            if isinstance(value, self.__class__):
                value = value.content

            if isinstance(value, dict):
                existing.update(value)
            else:
                raise ValueError(
                    "Cannot update a dictionary with a non-dictionary value."
                )

    @classmethod
    def from_dict(cls, kwargs: Any) -> "Note":
        """Create Note instance from dictionary.

        Args:
            kwargs: Dictionary to initialize with

        Returns:
            New Note instance
        """
        return cls(**kwargs)

    def __contains__(self, indices: INDICE_TYPE) -> bool:
        """Check if indices exist in content.

        Args:
            indices: Path to check

        Returns:
            True if path exists, False otherwise
        """
        return self.content.get(indices, UNDEFINED) is not UNDEFINED

    def __len__(self) -> int:
        """Get length of content.

        Returns:
            Number of top-level keys
        """
        return len(self.content)

    def __iter__(self) -> Iterator[str]:
        """Get iterator over content.

        Returns:
            Iterator over top-level keys
        """
        return iter(self.content)

    def __next__(self) -> str:
        """Get next item from content iterator.

        Returns:
            Next key in iteration
        """
        return next(iter(self.content))

    @override
    def __str__(self) -> str:
        """Get string representation of content.

        Returns:
            String representation of content dict
        """
        return str(self.content)

    @override
    def __repr__(self) -> str:
        """Get detailed string representation of content.

        Returns:
            Detailed string representation of content dict
        """
        return repr(self.content)

    def __getitem__(self, indices: INDICE_TYPE) -> Any:
        """Get item using index notation.

        Args:
            indices: Path to value

        Returns:
            Value at path

        Raises:
            KeyError: If path not found
        """
        indices = to_list(indices, flatten=True, dropna=True)
        return self.get(indices)

    def __setitem__(self, indices: INDICE_TYPE, value: Any) -> None:
        """Set item using index notation.

        Args:
            indices: Path where to set
            value: Value to set
        """
        self.set(indices, value)


__all__ = [
    "Note",
    "INDICE_TYPE",
]



from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    JsonValue,
    PrivateAttr,
    create_model,
    field_serializer,
    field_validator,
    model_serializer,
    model_validator,
)
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined

__all__ = [
    "BaseModel",
    "ConfigDict",
    "Field",
    "JsonValue",
    "PrivateAttr",
    "FieldInfo",
    "PydanticUndefined",
    "field_validator",
    "model_validator",
    "model_serializer",
    "field_serializer",
    "create_model",
]



from typing import List, Literal, Optional

from pydantic import Field

from ..data_models import OllamaEndpointRequestBody
from ..option_models import Option
from .message_models import Message
from .tool_models import Tool


class OllamaChatCompletionRequestBody(OllamaEndpointRequestBody):
    model: str = Field(description="The model name")

    messages: list[Message] = Field(
        description="The messages of the chat, this can be used to keep a chat memory"
    )

    tools: list[Tool] | None = Field(
        None,
        description="Tools for the model to use if supported. Requires 'stream' to be set to false",
    )

    format: Literal["json"] | None = Field(
        None,
        description="The format to return a response in. Currently the only accepted value is 'json'",
    )

    options: Option | None = Field(
        None,
        description="Additional model parameters listed in the documentation for the Modelfile",
    )

    stream: bool = Field(
        True,
        description="If false the response will be returned as a single response object, "
        "rather than a stream of objects",
    )

    keep_alive: str | Literal[0] = Field(
        "5m",
        description="Controls how long the model will stay loaded into memory following the request.",
    )



from .delete_file import OpenAIDeleteFilePathParam
from .list_files import OpenAIListFilesQueryParam
from .retrieve_file import OpenAIRetrieveFilePathParam
from .upload_file import OpenAIUploadFileRequestBody

__all__ = [
    "OpenAIUploadFileRequestBody",
    "OpenAIListFilesQueryParam",
    "OpenAIRetrieveFilePathParam",
    "OpenAIDeleteFilePathParam",
]



from typing import List, Optional

from pydantic import BaseModel, Field

from .function_models import ToolCall


class Message(BaseModel):
    content: str | None = Field(None, description="The contents of the message.")

    refusal: str | None = Field(
        None, description="The refusal message generated by the model."
    )

    tool_calls: list[ToolCall] | None = Field(
        None,
        description="The tool calls generated by the model, " "such as function calls.",
    )

    role: str | None = Field(
        None, description="The role of the author of this message."
    )



from typing import Any

import pandas as pd

from .to_df import to_df


def update_cells(
    df: pd.DataFrame,
    updates: dict[tuple[int | str, int | str], Any],
    create_missing: bool = False,
) -> pd.DataFrame:
    """
    Updates multiple cells in a DataFrame based on a dictionary of updates.

    Args:
        df: The DataFrame to update.
        updates: A dictionary where keys are (row, column) tuples and values are the new values.
                 Rows and columns can be specified by integer position or label.
        create_missing: If True, create new columns if they don't exist.

    Returns:
        The updated DataFrame.

    Raises:
        KeyError: If a specified row or column doesn't exist and create_missing is False.

    Example:
        >>> df = pd.DataFrame({'A': [1, 2], 'B': [3, 4]})
        >>> updates = {(0, 'A'): 10, (1, 'B'): 40, (0, 'C'): 100}
        >>> update_cells(df, updates, create_missing=True)
            A   B    C
        0  10   3  100
        1   2  40  NaN
    """
    df_copy = df.copy()

    for (row, col), value in updates.items():
        try:
            if isinstance(col, str) and col not in df_copy.columns:
                if create_missing:
                    df_copy[col] = None
                else:
                    raise KeyError(f"Column '{col}' does not exist.")

            df_copy.loc[row, col] = value
        except KeyError as e:
            if not create_missing:
                raise KeyError(f"Invalid row or column: {e}") from e

    return to_df(df_copy)



from typing import Optional

from pydantic import Field

from ..data_models import OllamaEndpointRequestBody, OllamaEndpointResponseBody


class OllamaPullModelRequestBody(OllamaEndpointRequestBody):
    name: str = Field(description="Name of the model to pull")

    insecure: bool = Field(
        False,
        description="Allow insecure connections to the library. "
        "Only use this if you are pulling from your own library during development.",
    )

    stream: bool = Field(
        True,
        description="If 'false' the response will be returned as a single response object, "
        "rather than a stream of objects",
    )


class OllamaPullModelResponseBody(OllamaEndpointResponseBody):
    status: str = Field(None)

    digest: str = Field(None)

    total: int = Field(None)

    completed: int = Field(None)



from typing import Any, Dict, List, Literal, Optional

from pydantic import BaseModel, Field


class FunctionParameters(BaseModel):
    type: Literal["object"] = "object"
    properties: dict[str, dict[str, Any]]
    required: list[str] | None = None


class Function(BaseModel):
    description: str | None = Field(
        None,
        description=(
            "A description of what the function does, used by the model to "
            "choose when and how to call the function."
        ),
    )

    name: str = Field(
        max_length=64,
        pattern="^[a-zA-Z0-9_-]+$",
        description=(
            "The name of the function to be called. Must be a-z, A-Z, 0-9, "
            "or contain underscores and dashes, with a maximum length of 64."
        ),
    )

    parameters: FunctionParameters | None = Field(
        None,
        description=(
            "The parameters the functions accepts, described as a JSON Schema "
            "object. See the guide for examples, and the JSON Schema "
            "reference for documentation about the format."
        ),
    )

    strict: bool | None = Field(
        False,
        description="Whether to enable strict schema"
        " adherence when generating the function call. "
        "If set to true, the model will follow the exact"
        " schema defined in the parameters field. "
        "Only a subset of JSON Schema is supported when strict is true.",
    )


class Tool(BaseModel):
    type: Literal["function"] = Field(
        description="The type of the tool. Currently," " only function is supported."
    )
    function: Function = Field(description="The function definition.")



from typing import List, Literal, Optional

from pydantic import Field

from ..data_models import OllamaEndpointRequestBody
from ..option_models import Option


class OllamaCompletionRequestBody(OllamaEndpointRequestBody):
    model: str = Field(description="The model name")

    prompt: str = Field(None, description="The prompt to generate a response for")

    suffix: str | None = Field(None, description="The text after the model response")

    images: list[str] | None = Field(
        None,
        description="A list of base64-encoded images (for multimodal models such as 'llava')",
    )

    format: Literal["json"] | None = Field(
        None,
        description="The format to return a response in. Currently the only accepted value is 'json'",
    )

    options: Option | dict | None = Field(
        None,
        description="Additional model parameters listed in the documentation for the 'Modelfile'",
    )

    system: str | None = Field(
        None,
        description="System message to (overrides what is defined in the 'Modelfile')",
    )

    template: str | None = Field(
        None,
        description="The prompt template to use (overrides what is defined in the 'Modelfile')",
    )

    context: list[int] | None = Field(
        None,
        description="The context parameter returned from a previous request to '/generate', "
        "this can be used to keep a short conversational memory",
    )

    stream: bool = Field(
        True,
        description="If 'false' the response will be returned as a single response object, "
        "rather than a stream of objects",
    )

    raw: bool = Field(
        False,
        description="if 'true' no formatting will be applied to the prompt. "
        "You may choose to use the 'raw' parameter if you are specifying a full templated "
        "prompt in your request to the API",
    )

    keep_alive: str | Literal[0] = Field(
        "5m",
        description="Controls how long the model will stay loaded into memory following the request.",
    )



import asyncio
from functools import wraps

from .rate_limiter import RateLimitError


def invoke_retry(max_retries: int = 3, base_delay: int = 1, max_delay: int = 60):
    def decorator(func):
        @wraps(func)
        async def wrapper(request_model, *args, **kwargs):
            if max_retries <= 0:
                raise ValueError(
                    "Invalid max number of retries. It must a positive integer."
                )

            for retry in range(max_retries + 1):
                try:
                    response_body = await func(request_model, *args, **kwargs)
                    return response_body
                except Exception as e:
                    # Last try used
                    if retry == max_retries:
                        raise e

                    # RateLimitError for Model only
                    if isinstance(e, RateLimitError):
                        if e.requested_tokens > request_model.rate_limiter.limit_tokens:
                            raise ValueError(
                                "Requested tokens exceed the model's token limit. "
                                "Please modify the input, adjust the expected output tokens, or increase the token limit. "
                                f"The current token limit is {request_model.rate_limiter.limit_tokens} tokens."
                            )

                        while request_model.rate_limiter.unreleased_requests:
                            await asyncio.sleep(2)
                            request_model.rate_limiter.release_tokens()
                            if request_model.rate_limiter.check_availability(
                                request_token_len=e.requested_tokens
                            ):
                                break

                    elif error_code := getattr(
                        e, "status", None
                    ):  # http request errors
                        if error_code == 429 and "exceeded your current quota" in str(
                            e
                        ):  # RateLimitError (account quota reached)
                            raise e
                        if error_code == 429 or error_code >= 500:  # ServerError
                            if retry_after := getattr(e, "headers", {}).get(
                                "Retry-After"
                            ):
                                if retry_after.isdigit():
                                    await asyncio.sleep(int(retry_after))
                                    continue

                            wait_time = min(base_delay * (2**retry), max_delay)
                            await asyncio.sleep(wait_time)
                    else:
                        raise e

        return wrapper

    return decorator



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from functools import singledispatchmethod

from lion.core.communication.action_request import ActionRequest
from lion.core.generic.log_manager import LogManager
from lion.core.typing import Any, Callable
from lion.libs.parse import to_dict, to_list
from lion.protocols.operatives.action import ActionRequestModel

from .function_calling import FunctionCalling
from .tool import Tool, func_to_tool

# Type definitions for tool registration and lookup
FUNCTOOL = Tool | Callable[..., Any]
FINDABLE_TOOL = FUNCTOOL | str
INPUTTABLE_TOOL = dict[str, Any] | bool | FINDABLE_TOOL
TOOL_TYPE = FINDABLE_TOOL | list[FINDABLE_TOOL] | INPUTTABLE_TOOL


class ActionManager:
    """Manages registration and execution of tools.

    The ActionManager serves as a central registry for tools, providing:
    1. Tool registration and management
    2. Function call matching to registered tools
    3. Tool invocation with argument handling
    4. Logging of tool execution

    Tools can be registered either as Tool objects or as callable functions
    (which are automatically converted to Tool objects). The manager supports
    various formats for function calls and provides comprehensive logging
    of tool execution.

    Attributes:
        registry (dict[str, Tool]): Dictionary mapping tool names to Tool objects.
        logger (LogManager): Logger for tracking tool execution.
    """

    def __init__(self, registry: dict[str, Tool] | None = None, logger=None) -> None:
        """Initialize the ActionManager instance.

        Args:
            registry: Optional dictionary of pre-registered tools.
                Keys are tool names, values are Tool objects.
            logger: Optional logger for tracking tool execution.
                If None, creates a new LogManager instance.
        """
        self.registry: dict[str, Tool] = registry or {}
        self.logger = logger or LogManager()

    def __contains__(self, tool: FINDABLE_TOOL) -> bool:
        """Check if a tool is registered in the registry.

        Supports checking by:
        - Tool object
        - Tool name (string)
        - Callable function

        Args:
            tool: The tool to check for registration.

        Returns:
            bool: True if tool is registered, False otherwise.
        """
        if isinstance(tool, Tool):
            return tool.function_name in self.registry
        elif isinstance(tool, str):
            return tool in self.registry
        elif callable(tool):
            return tool.__name__ in self.registry
        return False

    def register_tool(
        self,
        tool: FUNCTOOL,
        update: bool = False,
    ) -> None:
        """Register a single tool in the registry.

        If the tool is a callable function, it is automatically converted
        to a Tool object. Existing tools can be updated if update=True.

        Args:
            tool: The tool to register (Tool object or callable).
            update: If True, update existing tool; if False, raise error.

        Raises:
            ValueError: If tool already registered and update=False.
            TypeError: If tool is not a Tool object or callable.
        """
        if not update and tool in self:
            name = None
            if isinstance(tool, Tool):
                name = tool.function_name
            elif callable(tool):
                name = tool.__name__
            raise ValueError(f"Tool {name} is already registered.")

        if callable(tool):
            tool = func_to_tool(tool)[0]
        if not isinstance(tool, Tool):
            raise TypeError("Please register a Tool object or callable.")

        self.registry[tool.function_name] = tool

    def register_tools(
        self,
        tools: list[FUNCTOOL] | FUNCTOOL,
        update: bool = False,
    ) -> None:
        """Register multiple tools in the registry.

        Handles both single tools and lists of tools. Each tool can be
        either a Tool object or a callable function.

        Args:
            tools: Single tool or list of tools to register.
            update: If True, update existing tools; if False, raise error.

        Raises:
            ValueError: If any tool is already registered.
            TypeError: If any tool is not a Tool object or callable.
        """
        tools_list = tools if isinstance(tools, list) else [tools]
        [
            self.register_tool(tool, update=update)
            for tool in to_list(tools_list, dropna=True, flatten=True)
        ]

    @singledispatchmethod
    def match_tool(self, func_call: Any) -> FunctionCalling:
        """Match a function call to a registered tool.

        This is a single dispatch method that handles different function
        call formats. The base method handles unsupported types.

        Args:
            func_call: The function call specification.

        Raises:
            TypeError: If input type is not supported.
        """
        raise TypeError(f"Unsupported type {type(func_call)}")

    @match_tool.register
    def _(self, func_call: tuple) -> FunctionCalling:
        if len(func_call) == 2:
            function_name = func_call[0]
            arguments = func_call[1]
            tool = self.registry.get(function_name)
            if not tool:
                raise ValueError(f"Function {function_name} is not registered")
            return FunctionCalling(func_tool=tool, arguments=arguments)
        else:
            raise ValueError(f"Invalid function call {func_call}")

    @match_tool.register
    def _(self, func_call: dict) -> FunctionCalling:
        if len(func_call) == 2 and (
            {
                "function",
                "arguments",
            }
            <= func_call.keys()
        ):
            function_name = func_call["function"]
            tool = self.registry.get(function_name)
            if not tool:
                raise ValueError(f"Function {function_name} is not registered")
            return FunctionCalling(
                func_tool=tool,
                arguments=func_call["arguments"],
            )
        raise ValueError(f"Invalid function call {func_call}")

    @match_tool.register
    def _(self, func_call: ActionRequest | ActionRequestModel) -> FunctionCalling:
        tool = self.registry.get(func_call.function)
        if not tool:
            func_ = func_call.function
            raise ValueError(f"Function {func_} is not registered.")
        return FunctionCalling(func_tool=tool, arguments=func_call.arguments)

    @match_tool.register
    def _(self, func_call: str) -> FunctionCalling:
        _call = None
        try:
            _call = to_dict(func_call, str_type="json", fuzzy_parse=True)
        except Exception as e:
            raise ValueError(f"Invalid function call {func_call}") from e

        if isinstance(_call, dict):
            return self.match_tool(_call)
        raise ValueError(f"Invalid function call {func_call}")

    async def invoke(self, func_call: dict | str | ActionRequest) -> Any:
        """Invoke a tool based on the provided function call.

        1. Matches function call to registered tool
        2. Creates FunctionCalling instance
        3. Invokes function with arguments
        4. Logs execution details
        5. Returns result

        Args:
            func_call: Function call specification in supported format.
            log_manager: Optional logger for execution tracking.

        Returns:
            Result of tool invocation after processing pipeline.

        Raises:
            ValueError: If function not registered or call format invalid.
        """
        function_calling = self.match_tool(func_call)
        result = await function_calling.invoke()
        await self.logger.alog(function_calling.to_log())
        return result

    @property
    def schema_list(self) -> list[dict[str, Any]]:
        """List all tool schemas currently registered.

        Returns:
            List of OpenAI function schemas for all registered tools.
        """
        return [tool.schema_ for tool in self.registry.values()]

    def get_tool_schema(
        self,
        tools: TOOL_TYPE = False,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Retrieve the schema for specific tools or all tools.

        Args:
            tools: Specification of which tools to retrieve schemas for.
                If True, return all tools. If False, return empty dict.
                Can also be a specific tool or list of tools.
            **kwargs: Additional keyword arguments to include in output.

        Returns:
            Dictionary containing tool schemas and additional kwargs.

        Raises:
            ValueError: If a specified tool is not registered.
            TypeError: If an unsupported tool type is provided.
        """
        if isinstance(tools, bool):
            if tools:
                tool_kwarg = {"tools": self.schema_list}
            else:
                tool_kwarg = {}
        else:
            tool_kwarg = {"tools": self._get_tool_schema(tools)}
        return tool_kwarg | kwargs

    def _get_tool_schema(
        self,
        tool: Any,
    ) -> dict[str, Any] | list[dict[str, Any]]:
        if isinstance(tool, dict):
            return tool
        if isinstance(tool, Callable):
            name = tool.__name__
            if name in self.registry:
                return self.registry[name].schema_
            raise ValueError(f"Tool {name} is not registered.")

        elif isinstance(tool, Tool) or isinstance(tool, str):
            name = tool.function_name if isinstance(tool, Tool) else tool
            if name in self.registry:
                return self.registry[name].schema_
            raise ValueError(f"Tool {name} is not registered.")
        elif isinstance(tool, list):
            return [self._get_tool_schema(t) for t in tool]
        raise TypeError(f"Unsupported type {type(tool)}")


__all__ = ["ActionManager"]



"""JSON parsing and manipulation utilities."""

from .as_readable import as_readable, as_readable_json
from .extract import extract_block, extract_json_blocks
from .parse import _clean_json_string, fix_json_string, fuzzy_parse_json
from .schema import (
    extract_json_schema,
    json_schema_to_cfg,
    json_schema_to_regex,
    print_cfg,
)
from .to_json import to_json

__all__ = [
    "as_readable",
    "as_readable_json",
    "extract_block",
    "extract_json_blocks",
    "fuzzy_parse_json",
    "fix_json_string",
    "_clean_json_string",
    "extract_json_schema",
    "json_schema_to_cfg",
    "json_schema_to_regex",
    "print_cfg",
    "to_json",
]



def match_data_model(task_name):
    if task_name == "create_speech":
        from lion.integrations.openai_.api_endpoints.audio.speech_models import (
            OpenAISpeechRequestBody,
        )

        return {"request_body": OpenAISpeechRequestBody}
    elif task_name == "create_transcription":
        from lion.integrations.openai_.api_endpoints.audio.transcription_models import (
            OpenAITranscriptionRequestBody,
        )

        return {"request_body": OpenAITranscriptionRequestBody}
    elif task_name == "create_translation":
        from lion.integrations.openai_.api_endpoints.audio.translation_models import (
            OpenAITranslationRequestBody,
        )

        return {"request_body": OpenAITranslationRequestBody}
    elif task_name == "create_chat_completion":
        from lion.integrations.openai_.api_endpoints.chat_completions.request.request_body import (
            OpenAIChatCompletionRequestBody,
        )

        return {"request_body": OpenAIChatCompletionRequestBody}
    elif task_name == "create_embeddings":
        from lion.integrations.openai_.api_endpoints.embeddings.request_body import (
            OpenAIEmbeddingRequestBody,
        )

        return {"request_body": OpenAIEmbeddingRequestBody}
    elif task_name == "create_fine_tuning_job":
        from lion.integrations.openai_.api_endpoints.fine_tuning.create_jobs import (
            OpenAICreateFineTuningJobRequestBody,
        )

        return {"json_data": OpenAICreateFineTuningJobRequestBody}
    elif task_name == "list_fine_tuning_jobs":
        from lion.integrations.openai_.api_endpoints.fine_tuning.list_fine_tuning_jobs import (
            OpenAIListFineTuningJobsQueryParam,
        )

        return {"params": OpenAIListFineTuningJobsQueryParam}
    elif task_name == "list_fine_tuning_events":
        from lion.integrations.openai_.api_endpoints.fine_tuning.list_fine_tuning_events import (
            OpenAIListFineTuningEventsPathParam,
            OpenAIListFineTuningEventsQueryParam,
        )

        return {
            "params": OpenAIListFineTuningEventsQueryParam,
            "path_param": OpenAIListFineTuningEventsPathParam,
        }
    elif task_name == "list_fine_tuning_checkpoints":
        from lion.integrations.openai_.api_endpoints.fine_tuning.list_fine_tuning_checkpoints import (
            OpenAIListFineTuningCheckpointsPathParam,
            OpenAIListFineTuningCheckpointsQueryParam,
        )

        return {
            "params": OpenAIListFineTuningCheckpointsQueryParam,
            "path_param": OpenAIListFineTuningCheckpointsPathParam,
        }
    elif task_name == "retrieve_fine_tuning_job":
        from lion.integrations.openai_.api_endpoints.fine_tuning.retrieve_jobs import (
            OpenAIRetrieveFineTuningJobPathParam,
        )

        return {"path_param": OpenAIRetrieveFineTuningJobPathParam}
    elif task_name == "cancel_fine_tuning":
        from lion.integrations.openai_.api_endpoints.fine_tuning.cancel_jobs import (
            OpenAICancelFineTuningPathParam,
        )

        return {"path_param": OpenAICancelFineTuningPathParam}
    elif task_name == "create_batch":
        from lion.integrations.openai_.api_endpoints.batch.create_batch import (
            OpenAIBatchRequestBody,
        )

        return {"json_data": OpenAIBatchRequestBody}
    elif task_name == "retrieve_batch":
        from lion.integrations.openai_.api_endpoints.batch.retrieve_batch import (
            OpenAIRetrieveBatchPathParam,
        )

        return {"path_param": OpenAIRetrieveBatchPathParam}
    elif task_name == "cancel_batch":
        from lion.integrations.openai_.api_endpoints.batch.cancel_batch import (
            OpenAICancelBatchPathParam,
        )

        return {"path_param": OpenAICancelBatchPathParam}
    elif task_name == "list_batch":
        from lion.integrations.openai_.api_endpoints.batch.list_batch import (
            OpenAIListBatchQueryParam,
        )

        return {"params": OpenAIListBatchQueryParam}
    elif task_name == "upload_file":
        from lion.integrations.openai_.api_endpoints.files.upload_file import (
            OpenAIUploadFileRequestBody,
        )

        return {"form_data": OpenAIUploadFileRequestBody}
    elif task_name == "list_files":
        from lion.integrations.openai_.api_endpoints.files.list_files import (
            OpenAIListFilesQueryParam,
        )

        return {"params": OpenAIListFilesQueryParam}
    elif task_name == "retrieve_file" or task_name == "retrieve_file_content":
        from lion.integrations.openai_.api_endpoints.files.retrieve_file import (
            OpenAIRetrieveFilePathParam,
        )

        return {"path_param": OpenAIRetrieveFilePathParam}
    elif task_name == "delete_file":
        from lion.integrations.openai_.api_endpoints.files.delete_file import (
            OpenAIDeleteFilePathParam,
        )

        return {"path_param": OpenAIDeleteFilePathParam}
    elif task_name == "create_upload":
        from lion.integrations.openai_.api_endpoints.uploads.create_upload import (
            OpenAIUploadRequestBody,
        )

        return {"json_data": OpenAIUploadRequestBody}
    elif task_name == "add_upload_part":
        from lion.integrations.openai_.api_endpoints.uploads.add_upload_part import (
            OpenAIUploadPartPathParam,
            OpenAIUploadPartRequestBody,
        )

        return {
            "path_param": OpenAIUploadPartPathParam,
            "form_data": OpenAIUploadPartRequestBody,
        }
    elif task_name == "complete_upload":
        from lion.integrations.openai_.api_endpoints.uploads.complete_upload import (
            OpenAICompleteUploadPathParam,
            OpenAICompleteUploadRequestBody,
        )

        return {
            "path_param": OpenAICompleteUploadPathParam,
            "json_data": OpenAICompleteUploadRequestBody,
        }
    elif task_name == "cancel_upload":
        from lion.integrations.openai_.api_endpoints.uploads.cancel_upload import (
            OpenAICancelUploadPathParam,
        )

        return {"path_param": OpenAICancelUploadPathParam}
    elif task_name == "create_image":
        from lion.integrations.openai_.api_endpoints.images.image_models import (
            OpenAIImageRequestBody,
        )

        return {"json_data": OpenAIImageRequestBody}
    elif task_name == "create_image_edit":
        from lion.integrations.openai_.api_endpoints.images.image_edit_models import (
            OpenAIImageEditRequestBody,
        )

        return {"form_data": OpenAIImageEditRequestBody}
    elif task_name == "create_image_variation":
        from lion.integrations.openai_.api_endpoints.images.image_variation_models import (
            OpenAIImageVariationRequestBody,
        )

        return {"form_data": OpenAIImageVariationRequestBody}
    elif task_name == "list_models":
        return {}
    elif task_name == "retrieve_model":
        from lion.integrations.openai_.api_endpoints.models.retrieve_model import (
            OpenAIRetrieveModelPathParam,
        )

        return {"path_param": OpenAIRetrieveModelPathParam}
    elif task_name == "delete_fine_tuned_model":
        from lion.integrations.openai_.api_endpoints.models.delete_fine_tuned_model import (
            OpenAIDeleteFineTunedModelPathParam,
        )

        return {"path_param": OpenAIDeleteFineTunedModelPathParam}
    elif task_name == "create_moderation":
        from lion.integrations.openai_.api_endpoints.moderations.request_body import (
            OpenAIModerationRequestBody,
        )

        return {"json_data": OpenAIModerationRequestBody}
    else:
        raise ValueError(f"Invalid task: {task_name}. Not supported in the service.")



"""
Copyright 2024 HaiyangLi

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

from lion.core.generic.element import Element
from lion.core.typing import Any, Field, Note, PrivateAttr, Self, field_serializer
from lion.libs.parse import to_dict


class Log(Element):
    """A log entry in the Lion framework.

    The Log class represents an immutable log entry with structured content
    and metadata. Once created and loaded from storage, logs become immutable
    to ensure data integrity.

    Key Features:
        - Structured content storage using Note objects
        - Additional metadata through loginfo
        - Immutability after loading from storage
        - Serialization support
        - Validation of content and metadata

    Attributes:
        content (Note): The main content of the log entry
        loginfo (Note): Additional metadata about the log entry
    """

    content: Note = Field(
        default_factory=Note,
        title="Log Content",
        description="The content of the log entry.",
    )

    loginfo: Note = Field(
        default_factory=Note,
        title="Log Info",
        description="Metadata about the log entry.",
    )

    _immutable: bool = PrivateAttr(False)

    def __init__(self, content: Note, loginfo: Note, **kwargs) -> None:
        """Initialize a Log instance.

        Args:
            content: The main content of the log entry
            loginfo: Additional metadata about the log entry
            **kwargs: Additional arguments passed to Element
        """
        super().__init__(**kwargs)
        self.content = self._validate_note(content)
        self.loginfo = self._validate_note(loginfo)

    @classmethod
    def _validate_load_data(cls, data: dict, /) -> dict:
        try:
            # Validate required fields
            if not isinstance(data.get("content"), (dict, Note)):
                raise ValueError("Missing or invalid 'content' field")
            if not isinstance(data.get("loginfo"), (dict, Note)):
                raise ValueError("Missing or invalid 'loginfo' field")

            # Convert log_* fields to standard fields
            if "log_id" in data:
                data["ln_id"] = data.pop("log_id")
            if "log_timestamp" in data:
                data["timestamp"] = data.pop("log_timestamp")
            if "log_class" in data:
                data["lion_class"] = data.pop("log_class")
            return data
        except Exception as e:
            raise ValueError(
                "Log can only be loaded from a previously saved log entries.",
            ) from e

    @classmethod
    def from_dict(cls, data: dict, /) -> Self:
        """Create a Log instance from a dictionary.

        The created log will be immutable to ensure data integrity.

        Args:
            data: Dictionary containing log data

        Returns:
            New immutable Log instance
        """
        data = cls._validate_load_data(data)
        # Convert content and loginfo to Note objects if they're dicts
        if isinstance(data.get("content"), dict):
            data["content"] = Note(**data["content"])
        if isinstance(data.get("loginfo"), dict):
            data["loginfo"] = Note(**data["loginfo"])
        # Remove lion_class as it's handled by Element
        data.pop("lion_class", None)
        self = cls(**data)
        self._immutable = True
        return self

    def __setattr__(self, name: str, value: Any, /) -> None:
        if hasattr(self, "_immutable") and self._immutable:
            raise AttributeError("Cannot modify immutable log entry.")
        super().__setattr__(name, value)

    def _validate_note(self, value: Any, /) -> Note:

        if not value:
            return Note()
        if isinstance(value, Note):
            return value
        if isinstance(value, dict):
            return Note(**value)
        try:
            return Note(**to_dict(value))
        except Exception as e:
            raise ValueError(f"Invalid note value: {value}") from e

    @field_serializer("content", "loginfo")
    def _serialize_note(self, value: Note) -> dict:
        return value.to_dict()

    def to_dict(self) -> dict:
        """Convert Log to dictionary.

        Converts standard field names to log_* format and handles recursive
        serialization of nested structures.

        Returns:
            Dictionary representation of the log
        """
        dict_ = super().to_dict()
        # Convert standard fields to log_* fields
        dict_["log_id"] = dict_.pop("ln_id")
        dict_["log_class"] = dict_.pop("lion_class")
        dict_["log_timestamp"] = dict_.pop("timestamp")

        dict_ = to_dict(
            dict_,
            recursive=True,
            recursive_python_only=False,
            max_recursive_depth=10,  # Increased depth for nested structures
        )
        return dict_



"""
Copyright 2024 HaiyangLi

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

from inspect import isclass
from typing import Any, TypeVar, get_args, get_origin

from pydantic import BaseModel

T = TypeVar("T", bound=BaseModel)


def break_down_pydantic_annotation(
    model: type[T], max_depth: int | None = None, current_depth: int = 0
) -> dict[str, Any]:
    """
    Break down the type annotations of a Pydantic model into a dictionary.

    This function recursively processes Pydantic models, converting their
    field annotations into a dictionary structure. It handles nested models
    and lists of models.

    Args:
        model: The Pydantic model class to break down.
        max_depth: Maximum depth for recursion. None for no limit.
        current_depth: Current recursion depth (used internally).

    Returns:
        A dictionary representing the structure of the model's annotations.

    Raises:
        TypeError: If the input is not a Pydantic model.
        RecursionError: If max recursion depth is reached.

    Example:
        >>> from pydantic import BaseModel
        >>> class SubModel(BaseModel):
        ...     field1: int
        ...     field2: str
        >>> class MainModel(BaseModel):
        ...     sub: SubModel
        ...     items: list[SubModel]
        >>> result = break_down_annotation(MainModel)
        >>> print(result)
        {
            'sub': {'field1': <class 'int'>, 'field2': <class 'str'>},
            'items': [{'field1': <class 'int'>, 'field2': <class 'str'>}]
        }
    """

    if not _is_pydantic_model(model):
        raise TypeError("Input must be a Pydantic model")

    if max_depth is not None and current_depth >= max_depth:
        raise RecursionError("Maximum recursion depth reached")

    out: dict[str, Any] = {}
    for k, v in model.__annotations__.items():
        origin = get_origin(v)
        if _is_pydantic_model(v):
            out[k] = break_down_pydantic_annotation(v, max_depth, current_depth + 1)
        elif origin is list:
            args = get_args(v)
            if args and _is_pydantic_model(args[0]):
                out[k] = [
                    break_down_pydantic_annotation(
                        args[0], max_depth, current_depth + 1
                    )
                ]
            else:
                out[k] = [args[0] if args else Any]
        else:
            out[k] = v

    return out


def _is_pydantic_model(x: Any) -> bool:
    return isclass(x) and issubclass(x, BaseModel)



import ast
import copy as _copy
import importlib
import importlib.metadata
import importlib.util
import os
import random
import subprocess
import sys
from collections.abc import Mapping, Sequence
from datetime import datetime, timezone
from hashlib import sha256
from typing import Literal, TypeVar

T = TypeVar("T")


def unique_hash(n: int = 32) -> str:
    """unique random hash"""
    current_time = datetime.now().isoformat().encode("utf-8")
    random_bytes = os.urandom(42)
    return sha256(current_time + random_bytes).hexdigest()[:n]


def is_same_dtype(
    input_: list | dict, dtype: type | None = None, return_dtype: bool = False
) -> bool | tuple[bool, type]:
    """Check if all elements in input have the same data type."""
    if not input_:
        return True if not return_dtype else (True, None)

    iterable = input_.values() if isinstance(input_, Mapping) else input_
    first_element_type = type(next(iter(iterable), None))

    dtype = dtype or first_element_type
    result = all(isinstance(element, dtype) for element in iterable)
    return (result, dtype) if return_dtype else result


def insert_random_hyphens(
    s: str,
    num_hyphens: int = 1,
    start_index: int | None = None,
    end_index: int | None = None,
) -> str:
    """Insert random hyphens into a string."""
    if len(s) < 2:
        return s

    prefix = s[:start_index] if start_index else ""
    postfix = s[end_index:] if end_index else ""
    modifiable_part = s[start_index:end_index] if start_index else s

    positions = random.sample(range(len(modifiable_part)), num_hyphens)
    positions.sort()

    for pos in reversed(positions):
        modifiable_part = modifiable_part[:pos] + "-" + modifiable_part[pos:]

    return prefix + modifiable_part + postfix


def get_file_classes(file_path):
    with open(file_path) as file:
        file_content = file.read()

    tree = ast.parse(file_content)

    class_file_dict = {}
    for node in tree.body:
        if isinstance(node, ast.ClassDef):
            class_file_dict[node.name] = file_path

    return class_file_dict


def get_class_file_registry(folder_path, pattern_list):
    class_file_registry = {}
    for root, _, files in os.walk(folder_path):
        for file in files:
            if file.endswith(".py"):
                if any(pattern in root for pattern in pattern_list):
                    class_file_dict = get_file_classes(os.path.join(root, file))
                    class_file_registry.update(class_file_dict)
    return class_file_registry


def get_class_objects(file_path):
    class_objects = {}
    spec = importlib.util.spec_from_file_location("module.name", file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    for class_name in dir(module):
        obj = getattr(module, class_name)
        if isinstance(obj, type):
            class_objects[class_name] = obj

    return class_objects


def time(
    *,
    tz: timezone = timezone.utc,
    type_: Literal["timestamp", "datetime", "iso", "custom"] = "timestamp",
    sep: str | None = "T",
    timespec: str | None = "auto",
    custom_format: str | None = None,
    custom_sep: str | None = None,
) -> float | str | datetime:
    """
    Get current time in various formats.

    Args:
        tz: Timezone for the time (default: utc).
        type_: Type of time to return (default: "timestamp").
            Options: "timestamp", "datetime", "iso", "custom".
        sep: Separator for ISO format (default: "T").
        timespec: Timespec for ISO format (default: "auto").
        custom_format: Custom strftime format string for
            type_="custom".
        custom_sep: Custom separator for type_="custom",
            replaces "-", ":", ".".

    Returns:
        Current time in the specified format.

    Raises:
        ValueError: If an invalid type_ is provided or if custom_format
            is not provided when type_="custom".
    """
    now = datetime.now(tz=tz)

    if type_ == "iso":
        return now.isoformat(sep=sep, timespec=timespec)
    elif type_ == "timestamp":
        return now.timestamp()
    elif type_ == "datetime":
        return now
    elif type_ == "custom":
        if not custom_format:
            raise ValueError("custom_format must be provided when type_='custom'")
        formatted_time = now.strftime(custom_format)
        if custom_sep is not None:
            for old_sep in ("-", ":", "."):
                formatted_time = formatted_time.replace(old_sep, custom_sep)
        return formatted_time

    raise ValueError(
        f"Invalid value <{type_}> for `type_`, must be"
        " one of 'timestamp', 'datetime', 'iso', or 'custom'."
    )


def copy(obj: T, /, *, deep: bool = True, num: int = 1) -> T | list[T]:
    """
    Create one or more copies of an object.

    Args:
        obj: The object to be copied.
        deep: If True, create a deep copy. Otherwise, create a shallow
            copy.
        num: The number of copies to create.

    Returns:
        A single copy if num is 1, otherwise a list of copies.

    Raises:
        ValueError: If num is less than 1.
    """
    if num < 1:
        raise ValueError("Number of copies must be at least 1")

    copy_func = _copy.deepcopy if deep else _copy.copy
    return [copy_func(obj) for _ in range(num)] if num > 1 else copy_func(obj)


def run_pip_command(
    args: Sequence[str],
) -> subprocess.CompletedProcess[bytes]:
    """Run a pip command."""
    return subprocess.run(
        [sys.executable, "-m", "pip"] + list(args),
        check=True,
        capture_output=True,
    )


def format_deprecation_msg(
    deprecated_name: str,
    type_: str,
    deprecated_version: str,
    removal_version: str,
    replacement: str | Literal["python"] | None = None,
    additional_msg: str | None = None,
) -> None:

    msg = (
        f"{type_}: <{deprecated_name}> is deprecated since "
        f"<{deprecated_version}> and will be removed in {removal_version}."
    )
    if replacement is None:
        msg += " No replacement is available."
    elif replacement == "python":
        msg += " Use the Python standard library instead."
    elif replacement:
        msg += f" Use <{replacement}> instead."
    if additional_msg:
        msg += f" {additional_msg}"
    return msg


def get_bins(input_: list[str], upper: int) -> list[list[int]]:
    """Organizes indices of strings into bins based on a cumulative upper limit.

    Args:
        input_ (List[str]): The list of strings to be binned.
        upper (int): The cumulative length upper limit for each bin.

    Returns:
        List[List[int]]: A list of bins, each bin is a list of indices from the input list.
    """
    current = 0
    bins = []
    current_bin = []
    for idx, item in enumerate(input_):
        if current + len(item) < upper:
            current_bin.append(idx)
            current += len(item)
        else:
            bins.append(current_bin)
            current_bin = [idx]
            current = len(item)
    if current_bin:
        bins.append(current_bin)
    return bins


__all__ = [
    "unique_hash",
    "is_same_dtype",
    "insert_random_hyphens",
    "get_file_classes",
    "get_class_file_registry",
    "get_class_objects",
    "time",
    "copy",
    "run_pip_command",
    "format_deprecation_msg",
    "get_bins",
]



from abc import ABC
from collections.abc import (
    Callable,
    Container,
    ItemsView,
    Iterator,
    Mapping,
    Sequence,
    ValuesView,
)
from enum import Enum
from typing import (
    Annotated,
    Any,
    ClassVar,
    Generic,
    Literal,
    NoReturn,
    Self,
    TypeAlias,
    TypeVar,
)

from typing_extensions import override

from lion.libs.constants import UNDEFINED, UndefinedType

__all__ = [
    "ABC",
    "Annotated",
    "Any",
    "ClassVar",
    "Container",
    "Enum",
    "Generic",
    "Literal",
    "Mapping",
    "NoReturn",
    "Sequence",
    "TypeAlias",
    "TypeVar",
    "UndefinedType",
    "UNDEFINED",
    "override",
    "Callable",
    "Self",
    "Iterator",
    "ValuesView",
    "ItemsView",
]



import json
from typing import Any

from ..type_convert.to_dict import to_dict


def as_readable_json(input_: Any, /, **kwargs) -> str:
    """Convert input to a human-readable JSON string.

    Args:
        input_: Object to convert to readable JSON
        **kwargs: Additional arguments passed to json.dumps()

    Returns:
        A formatted, human-readable JSON string

    Raises:
        ValueError: If conversion to JSON fails
    """
    # Extract to_dict kwargs
    to_dict_kwargs = {
        "use_model_dump": True,
        "fuzzy_parse": True,
        "recursive": True,
        "recursive_python_only": True,
        "max_recursive_depth": 5,
    }
    to_dict_kwargs.update(kwargs)

    # Handle empty input
    if not input_:
        if isinstance(input_, list):
            return ""
        return "{}"

    try:
        if isinstance(input_, list):
            # For lists, convert and format each item separately
            items = []
            for item in input_:
                dict_ = to_dict(item, **to_dict_kwargs)
                items.append(
                    json.dumps(
                        dict_,
                        indent=4,
                        ensure_ascii=False,
                        default=lambda o: to_dict(o),
                    )
                )
            return "\n\n".join(items)

        # Handle single items
        dict_ = to_dict(input_, **to_dict_kwargs)

        # Extract json.dumps kwargs from input kwargs
        json_kwargs = {
            "indent": 4,
            "ensure_ascii": kwargs.get("ensure_ascii", False),
            "default": lambda o: to_dict(o),
        }

        # Add any other JSON-specific kwargs
        for k in ["indent", "separators", "cls"]:
            if k in kwargs:
                json_kwargs[k] = kwargs[k]

        # Convert to JSON string
        if kwargs.get("ensure_ascii", False):
            # Force ASCII encoding for special characters
            return json.dumps(
                dict_,
                ensure_ascii=True,
                **{k: v for k, v in json_kwargs.items() if k != "ensure_ascii"},
            )

        return json.dumps(dict_, **json_kwargs)

    except Exception as e:
        raise ValueError(f"Failed to convert input to readable JSON: {e}") from e


def as_readable(input_: Any, /, *, md: bool = False, **kwargs) -> str:
    """Convert input to readable string with optional markdown formatting.

    Args:
        input_: Object to convert
        md: Whether to wrap in markdown block
        **kwargs: Additional arguments for as_readable_json()

    Returns:
        Formatted string representation
    """
    try:
        result = as_readable_json(input_, **kwargs)
        if md:
            return f"```json\n{result}\n```"
        return result

    except Exception:
        return str(input_)



from pydantic import Field

from ..data_models import OpenAIEndpointRequestBody


class OpenAIUploadRequestBody(OpenAIEndpointRequestBody):
    filename: str = Field(description="The name of the file to upload.")

    purpose: str = Field(description="The intended purpose of the uploaded file.")

    bytes: int = Field(description="The intended number of bytes to be uploaded.")

    mime_type: str = Field(description="The MIME type of the file.")



import inspect

from lion.libs.parse import validate_boolean
from lion.libs.utils import copy

from ..typing.pydantic_ import (
    BaseModel,
    Field,
    FieldInfo,
    PrivateAttr,
    create_model,
    field_validator,
    model_validator,
)
from ..typing.typing_ import Callable, Self
from .field_model import FieldModel
from .schema_model import SchemaModel


class NewModelParams(SchemaModel):
    """Configuration class for dynamically creating new Pydantic models."""

    name: str | None = None
    parameter_fields: dict[str, FieldInfo] = Field(default_factory=dict)
    base_type: type[BaseModel] = Field(default=BaseModel)
    field_models: list[FieldModel] = Field(default_factory=list)
    exclude_fields: list = Field(default_factory=list)
    field_descriptions: dict = Field(default_factory=dict)
    inherit_base: bool = Field(default=True)
    config_dict: dict | None = Field(default=None)
    doc: str | None = Field(default=None)
    frozen: bool = False
    _validators: dict[str, Callable] | None = PrivateAttr(default=None)
    _use_keys: set[str] = PrivateAttr(default_factory=set)

    @property
    def use_fields(self):
        """Get field definitions to use in new model."""
        params = {k: v for k, v in self.parameter_fields.items() if k in self._use_keys}
        params.update(
            {
                f.name: f.field_info
                for f in self.field_models
                if f.name in self._use_keys
            }
        )
        return {k: (v.annotation, v) for k, v in params.items()}

    @field_validator("parameter_fields", mode="before")
    def validate_parameters(cls, value):
        """Validate parameter field definitions."""
        if value is None:
            return {}
        if not isinstance(value, dict):
            raise ValueError("Fields must be a dictionary.")
        for k, v in value.items():
            if not isinstance(k, str):
                raise ValueError("Field names must be strings.")
            if not isinstance(v, FieldInfo):
                raise ValueError("Field values must be FieldInfo objects.")
        return copy(value)

    @field_validator("base_type", mode="before")
    def validate_base(cls, value) -> type[BaseModel]:
        """Validate base model type."""
        if value is None:
            return BaseModel
        if isinstance(value, type) and issubclass(value, BaseModel):
            return value
        if isinstance(value, BaseModel):
            return value.__class__
        raise ValueError("Base must be a BaseModel subclass or instance.")

    @field_validator("exclude_fields", mode="before")
    def validate_fields(cls, value) -> list[str]:
        """Validate excluded fields list."""
        if value is None:
            return []
        if isinstance(value, dict):
            value = list(value.keys())
        if isinstance(value, set | tuple):
            value = list(value)
        if isinstance(value, list):
            if not all(isinstance(i, str) for i in value):
                raise ValueError("Field names must be strings.")
            return copy(value)
        raise ValueError("Fields must be a list, set, or dictionary.")

    @field_validator("field_descriptions", mode="before")
    def validate_field_descriptions(cls, value) -> dict[str, str]:
        """Validate field descriptions dictionary."""
        if value is None:
            return {}
        if not isinstance(value, dict):
            raise ValueError("Field descriptions must be a dictionary.")
        for k, v in value.items():
            if not isinstance(k, str):
                raise ValueError("Field names must be strings.")
            if not isinstance(v, str):
                raise ValueError("Field descriptions must be strings.")
        return value

    @field_validator("inherit_base", mode="before")
    def validate_inherit_base(cls, value) -> bool:
        """Validate inherit_base flag."""
        try:
            return validate_boolean(value)
        except Exception:
            return True

    @field_validator("name", mode="before")
    def validate_name(cls, value) -> str:
        """Validate model name."""
        if value is None:
            return "StepModel"
        if not isinstance(value, str):
            raise ValueError("Name must be a string.")
        return value

    @field_validator("field_models", mode="before")
    def _validate_field_models(cls, value):
        """Validate field model definitions."""
        if value is None:
            return []
        value = [value] if not isinstance(value, list) else value
        if not all(isinstance(i, FieldModel) for i in value):
            raise ValueError("Field models must be FieldModel objects.")
        return value

    @model_validator(mode="after")
    def validate_param_model(self) -> Self:
        """Validate complete model configuration."""
        if self.base_type is not None:
            self.parameter_fields.update(copy(self.base_type.model_fields))

        self.parameter_fields.update({f.name: f.field_info for f in self.field_models})

        use_keys = list(self.parameter_fields.keys())
        use_keys.extend(list(self._use_keys))

        if self.exclude_fields:
            use_keys = [i for i in use_keys if i not in self.exclude_fields]

        self._use_keys = set(use_keys)

        validators = {}

        for i in self.field_models:
            if i.field_validator is not None:
                validators.update(i.field_validator)
        self._validators = validators

        if self.field_descriptions:
            for i in self.field_models:
                if i.name in self.field_descriptions:
                    i.description = self.field_descriptions[i.name]

        if not isinstance(self.name, str):
            if hasattr(self.base_type, "class_name"):
                if callable(self.base_type.class_name):
                    self.name = self.base_type.class_name()
                else:
                    self.name = self.base_type.class_name
            elif inspect.isclass(self.base_type):
                self.name = self.base_type.__name__

        return self

    def create_new_model(self) -> type[BaseModel]:
        """Create new Pydantic model with specified configuration."""
        a: type[BaseModel] = create_model(
            self.name,
            __config__=self.config_dict,
            __doc__=self.doc,
            __base__=self.base_type if self.inherit_base else None,
            __validators__=self._validators,
            **self.use_fields,
        )
        if self.frozen:
            a.model_config["frozen"] = True
        return a


__all__ = ["NewModelParams"]



import inspect

from dotenv import load_dotenv

from lion.integrations.anthropic_.api_endpoints.api_request import AnthropicRequest
from lion.integrations.anthropic_.api_endpoints.messages.request.request_body import (
    AnthropicMessageRequestBody,
)
from lion.service import Service, register_service

from .AnthropicModel import AnthropicModel

load_dotenv()


@register_service
class AnthropicService(Service):
    def __init__(
        self,
        api_key: str,
        api_version: str = "2023-06-01",
        name: str = None,
    ):
        super().__setattr__("_initialized", False)
        self.api_key = api_key
        self.api_version = api_version
        self.name = name
        self.rate_limiters = {}  # model: RateLimiter
        super().__setattr__("_initialized", True)

    def __setattr__(self, key, value):
        if getattr(self, "_initialized", False) and key in [
            "api_key",
            "api_version",
        ]:
            raise AttributeError(
                f"Cannot modify '{key}' after initialization. "
                f"Please set a new service object for new keys."
            )
        super().__setattr__(key, value)

    def check_rate_limiter(
        self,
        anthropic_model: AnthropicModel,
        limit_requests: int = None,
        limit_tokens: int = None,
    ):
        # Map model versions to their base models for shared rate limiting
        shared_models = {
            "claude-3-opus-20240229": "claude-3-opus",
            "claude-3-sonnet-20240229": "claude-3-sonnet",
            "claude-3-haiku-20240307": "claude-3-haiku",
        }

        if anthropic_model.model in shared_models:
            model = shared_models[anthropic_model.model]
        else:
            model = anthropic_model.model

        if model not in self.rate_limiters:
            self.rate_limiters[model] = anthropic_model.rate_limiter
        else:
            anthropic_model.rate_limiter = self.rate_limiters[model]
            if limit_requests:
                anthropic_model.rate_limiter.limit_requests = limit_requests
            if limit_tokens:
                anthropic_model.rate_limiter.limit_tokens = limit_tokens

        return anthropic_model

    @staticmethod
    def match_data_model(task_name: str) -> dict:
        """Match task name to appropriate request and response models."""
        if task_name == "create_message":
            return {"request_body": AnthropicMessageRequestBody}
        raise ValueError(f"No data models found for task: {task_name}")

    @classmethod
    def list_tasks(cls):
        methods = []
        for name, member in inspect.getmembers(cls, predicate=inspect.isfunction):
            if name not in [
                "__init__",
                "__setattr__",
                "check_rate_limiter",
                "match_data_model",
            ]:
                methods.append(name)
        return methods

    # Messages
    def create_message(
        self, model: str, limit_tokens: int = None, limit_requests: int = None
    ):
        model_obj = AnthropicModel(
            model=model,
            api_key=self.api_key,
            api_version=self.api_version,
            endpoint="messages",
            method="POST",
            content_type="application/json",
            limit_tokens=limit_tokens,
            limit_requests=limit_requests,
        )

        return self.check_rate_limiter(
            model_obj, limit_requests=limit_requests, limit_tokens=limit_tokens
        )



from typing import List

from pydantic import BaseModel, Field

from ..data_models import OllamaEndpointResponseBody


class Detail(BaseModel):
    parent_model: str = Field(None)

    format: str = Field(None)

    family: str = Field(None)

    families: list[str] | None = Field(None)

    parameter_size: str = Field(None)

    quantization_level: str = Field(None)


class LocalModel(BaseModel):
    name: str = Field(None)

    model: str = Field(None)

    modified_at: str = Field(None)

    size: int = Field(None)

    digest: str = Field(None)

    details: Detail = Field(None)


class OllamaListLocalModelsResponseBody(OllamaEndpointResponseBody):
    models: list[LocalModel] = Field(None)


class RunningModel(BaseModel):
    name: str = Field(None)

    model: str = Field(None)

    size: int = Field(None)

    digest: str = Field(None)

    details: Detail = Field(None)

    expires_at: str = Field(None)

    size_vram: int = Field(None)


class OllamaListRunningModelsResponseBody(OllamaEndpointResponseBody):
    models: list[RunningModel] = Field(None)



"""
Copyright 2024 HaiyangLi

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from lion.core.typing import ID, Any, Note, override
from lion.libs.utils import copy

from .action_request import ActionRequest
from .message import MessageFlag, MessageRole, RoledMessage


def prepare_action_response_content(
    action_request: ActionRequest,
    output: Any,
) -> Note:
    """
    Prepare the content for an action response.

    Creates a standardized Note object containing the response details and
    maintaining reference to the original request.

    Args:
        action_request: The original action request being responded to
        output: The result from executing the requested function

    Returns:
        Note: Formatted action response content
    """
    return Note(
        action_request_id=action_request.ln_id,
        action_response={
            "function": action_request.function,
            "arguments": action_request.arguments,
            "output": output,
        },
    )


class ActionResponse(RoledMessage):
    """
    Represents a response to an action request in the system.

    This class encapsulates the result of executing an action request,
    maintaining references to the original request and providing access
    to the execution results.

    Example:
        >>> request = ActionRequest(function="sum", arguments={"nums": [1,2,3]})
        >>> response = ActionResponse(action_request=request, output=6)
        >>> print(response.output)
        6
    """

    @override
    def __init__(
        self,
        action_request: ID[ActionRequest].Item,
        output: Any = None,
        protected_init_params: dict | None = None,
    ) -> None:
        """
        Initialize an ActionResponse instance.

        Args:
            action_request: The original action request to respond to
            output: The result from the function execution
            protected_init_params: Protected initialization parameters

        Raises:
            ValueError: If action_request is invalid or already responded to
        """
        message_flags = [
            action_request,
            output,
        ]

        if all(x == MessageFlag.MESSAGE_LOAD for x in message_flags):
            protected_init_params = protected_init_params or {}
            super().__init__(**protected_init_params)
            return

        if all(x == MessageFlag.MESSAGE_CLONE for x in message_flags):
            super().__init__(role=MessageRole.ASSISTANT)
            return

        super().__init__(
            role=MessageRole.ASSISTANT,
            recipient=action_request.sender,
            sender=action_request.recipient,
            content=prepare_action_response_content(
                action_request=action_request,
                output=output,
            ),
        )
        action_request.content["action_response_id"] = self.ln_id

    @property
    def function(self) -> str:
        """
        Get the function name from the action response.

        Returns:
            str: The name of the function that was executed
        """
        return copy(self.content.get(["action_response", "function"]))

    @property
    def arguments(self) -> dict[str, Any]:
        """
        Get the function arguments from the action response.

        Returns:
            dict[str, Any]: The arguments that were used
        """
        return copy(self.content.get(["action_response", "arguments"]))

    @property
    def output(self) -> Any:
        """
        Get the function output from the action response.

        Returns:
            Any: The result returned by the function
        """
        return self.content.get(["action_response", "output"])

    @property
    def response(self) -> dict[str, Any]:
        """
        Get the complete action response as a dictionary.

        Returns:
            dict[str, Any]: The response including function details and output
        """
        return copy(self.content.get("action_response", {}))

    @property
    def action_request_id(self) -> ID[ActionRequest].ID | None:
        """
        Get the ID of the corresponding action request.

        Returns:
            ID[ActionRequest].ID | None: The ID of the original request
        """
        return copy(self.content.get("action_request_id", None))

    @override
    def _format_content(self) -> dict[str, Any]:
        return {"role": self.role.value, "content": self.response}



"""
Copyright 2024 HaiyangLi

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

from typing import Self

from pydantic import BaseModel, Field, PrivateAttr, model_validator
from pydantic.fields import FieldInfo

from lion.core.models import FieldModel, NewModelParams, OperableModel
from lion.core.typing import UNDEFINED
from lion.libs.parse import to_json, validate_keys


class Operative(OperableModel):
    """Class representing an operative that handles request and response models for operations."""

    name: str | None = None

    request_params: NewModelParams | None = Field(default=None)
    request_type: type[BaseModel] | None = Field(default=None)

    response_params: NewModelParams | None = Field(default=None)
    response_type: type[BaseModel] | None = Field(default=None)
    response_model: OperableModel | None = Field(default=None)
    response_str_dict: dict | str | None = Field(default=None)

    auto_retry_parse: bool = True
    max_retries: int = 3
    _should_retry: bool = PrivateAttr(default=None)

    @model_validator(mode="after")
    def _validate(self) -> Self:
        """Validates the operative instance after initialization."""
        if self.request_type is None:
            self.request_type = self.request_params.create_new_model()
        if self.name is None:
            self.name = self.request_params.name or self.request_type.__name__
        return self

    def raise_validate_pydantic(self, text: str) -> None:
        """Validates and updates the response model using strict matching.

        Args:
            text (str): The text to validate and parse into the response model.

        Raises:
            Exception: If the validation fails.
        """
        d_ = to_json(text, fuzzy_parse=True)
        if isinstance(d_, list | tuple) and len(d_) == 1:
            d_ = d_[0]
        try:
            d_ = validate_keys(
                d_, self.request_type.model_fields, handle_unmatched="raise"
            )
            d_ = {k: v for k, v in d_.items() if v != UNDEFINED}
            self.response_model = self.request_type.model_validate(d_)
            self._should_retry = False
        except Exception:
            self.response_str_dict = d_
            self._should_retry = True

    def force_validate_pydantic(self, text: str):
        """Forcibly validates and updates the response model, allowing unmatched fields.

        Args:
            text (str): The text to validate and parse into the response model.
        """
        d_ = text
        try:
            d_ = to_json(text, fuzzy_parse=True)
            if isinstance(d_, list | tuple) and len(d_) == 1:
                d_ = d_[0]
            d_ = validate_keys(
                d_, self.request_type.model_fields, handle_unmatched="force"
            )
            d_ = {k: v for k, v in d_.items() if v != UNDEFINED}
            self.response_model = self.request_type.model_validate(d_)
            self._should_retry = False
        except Exception:
            self.response_str_dict = d_
            self.response_model = None
            self._should_retry = True

    def update_response_model(
        self, text: str | None = None, data: dict | None = None
    ) -> BaseModel | dict | str | None:
        """Updates the response model based on the provided text or data.

        Args:
            text (str, optional): The text to parse and validate.
            data (dict, optional): The data to update the response model with.

        Returns:
            BaseModel | dict | str | None: The updated response model or raw data.

        Raises:
            ValueError: If neither text nor data is provided.
        """
        if text is None and data is None:
            raise ValueError("Either text or data must be provided.")

        if text:
            self.response_str_dict = text
            try:
                self.raise_validate_pydantic(text)
            except Exception:
                self.force_validate_pydantic(text)

        if data and self.response_type:
            d_ = self.response_model.model_dump()
            d_.update(data)
            self.response_model = self.response_type.model_validate(d_)

        if not self.response_model and isinstance(self.response_str_dict, list):
            try:
                self.response_model = [
                    self.request_type.model_validate(d_)
                    for d_ in self.response_str_dict
                ]
            except Exception:
                pass

        return self.response_model or self.response_str_dict

    def create_response_type(
        self,
        response_params: NewModelParams | None = None,
        field_models: list[FieldModel] | None = None,
        parameter_fields: dict[str, FieldInfo] | None = None,
        exclude_fields: list[str] | None = None,
        field_descriptions: dict[str, str] | None = None,
        inherit_base: bool = True,
        config_dict: dict | None = None,
        doc: str | None = None,
        frozen: bool = False,
        validators: dict | None = None,
    ) -> None:
        """Creates a new response type based on the provided parameters.

        Args:
            response_params (NewModelParams, optional): Parameters for the new response model.
            field_models (list[FieldModel], optional): List of field models.
            parameter_fields (dict[str, FieldInfo], optional): Dictionary of parameter fields.
            exclude_fields (list, optional): List of fields to exclude.
            field_descriptions (dict, optional): Dictionary of field descriptions.
            inherit_base (bool, optional): Whether to inherit the base model.
            config_dict (dict | None, optional): Configuration dictionary.
            doc (str | None, optional): Documentation string.
            frozen (bool, optional): Whether the model is frozen.
            validators (dict, optional): Dictionary of validators.
        """
        self.response_params = response_params or NewModelParams(
            parameter_fields=parameter_fields,
            field_models=field_models,
            exclude_fields=exclude_fields,
            field_descriptions=field_descriptions,
            inherit_base=inherit_base,
            config_dict=config_dict,
            doc=doc,
            frozen=frozen,
            base_type=self.request_params.base_type,
        )
        if validators and isinstance(validators, dict):
            self.response_params._validators.update(validators)

        self.response_type = self.response_params.create_new_model()



from enum import Enum
from typing import Literal, Optional

from pydantic import Field

from ..data_models import OpenAIEndpointRequestBody


class Size(str, Enum):
    S256 = "256x256"
    S512 = "512x512"
    S1024 = "1024x1024"
    S1792_1024 = "1792x1024"
    S1024_1792 = "1024x1792"


class Style(str, Enum):
    VIVID = "vivid"
    NATURAL = "natural"


class OpenAIImageRequestBody(OpenAIEndpointRequestBody):
    prompt: str = Field(description="A text description of the desired image(s).")

    model: str = Field("dall-e-2", description="The model to use for image generation.")

    n: int | None = Field(
        1, description="The number of images to generate.", ge=1, le=10
    )

    quality: Literal["standard", "hd"] | None = Field(
        "standard",
        description="The quality of the image that will be generated.",
    )

    response_format: Literal["url", "b64_json"] | None = Field(
        "url",
        description="The format in which the generated images are returned.",
    )

    size: Size | None = Field(
        "1024x1024", description="The size of the generated images."
    )

    style: Style | None = Field(
        "vivid", description="The style of the generated images."
    )

    user: str | None = Field(
        None,
        description="A unique identifier representing your end-user, which can help OpenAI to monitor and detect abuse.",
    )



from .action_manager import ActionManager
from .base import ObservableAction
from .function_calling import FunctionCalling
from .tool import Tool, func_to_tool

__all__ = [
    "ObservableAction",
    "FunctionCalling",
    "Tool",
    "func_to_tool",
    "ActionManager",
]



from typing import Literal

from pydantic import BaseModel, Field


class TextContent(BaseModel):
    type: Literal["text"] = Field(description="Type of content block")
    text: str = Field(description="The text content")


class ToolUseContent(BaseModel):
    type: Literal["tool_use"] = Field(description="Type of content block")
    id: str = Field(description="Unique identifier for this tool use")
    name: str = Field(description="Name of the tool being called")
    input: dict = Field(description="Input parameters for the tool call")


class ToolResultContent(BaseModel):
    type: Literal["tool_result"] = Field(description="Type of content block")
    tool_use_id: str = Field(description="ID of the tool use this is a result for")
    content: str = Field(description="The result content from running the tool")


ContentBlock = TextContent | ToolUseContent | ToolResultContent
